from gettext import gettext as _
import time
import logging
from gettext import ngettext
from typing import List, Iterable, Optional

from blueman.bluez.Device import Device
from blueman.bluez.errors import BluezDBusException
from blueman.main.Builder import Builder
from blueman.bluemantyping import GSignals
from blueman.bluez.Adapter import Adapter
from blueman.bluez.obex.ObjectPush import ObjectPush
from blueman.bluez.obex.Manager import Manager
from blueman.bluez.obex.Client import Client
from blueman.bluez.obex.Transfer import Transfer
from blueman.Functions import format_bytes
from blueman.main.SpeedCalc import SpeedCalc
from blueman.gui.CommonUi import ErrorDialog

import gi
gi.require_version("Gtk", "3.0")
gi.require_version("Gdk", "3.0")
from gi.repository import Gdk, Gtk, GObject, GLib, Gio


class Sender(Gtk.Dialog):
    __gsignals__: GSignals = {
        'result': (GObject.SignalFlags.RUN_FIRST, None, (GObject.TYPE_BOOLEAN,)),
    }

    def __init__(self, device: Device, adapter_path: str, files: Iterable[str]) -> None:
        super().__init__(
            title=_("Bluetooth File Transfer"),
            name="BluemanSendTo",
            icon_name="blueman",
            border_width=5,
            default_width=400,
            window_position=Gtk.WindowPosition.CENTER,
            type_hint=Gdk.WindowTypeHint.DIALOG
        )

        self.b_cancel = self.add_button("_Stop", Gtk.ResponseType.CLOSE)
        self.b_cancel.props.receives_default = True
        self.b_cancel.props.use_underline = True
        self.b_cancel.connect("clicked", self.on_cancel)

        builder = Builder("send-dialog.ui")

        grid = builder.get_widget("sendto", Gtk.Grid)
        content_area = self.get_content_area()
        content_area.add(grid)

        self.l_dest = builder.get_widget("l_dest", Gtk.Label)
        self.l_file = builder.get_widget("l_file", Gtk.Label)

        self.pb = builder.get_widget("pb", Gtk.ProgressBar)
        self.pb.props.text = _("Connecting")

        self.device = device
        self.adapter = Adapter(obj_path=adapter_path)
        self.manager = Manager()
        self.files: List[Gio.File] = []
        self.num_files = 0
        self.object_push: Optional[ObjectPush] = None
        self.object_push_handlers: List[int] = []
        self.transfer: Optional[Transfer] = None

        self.total_bytes = 0
        self.total_transferred = 0

        self._last_bytes = 0
        self._last_update = 0.0

        self.error_dialog: Optional[ErrorDialog] = None
        self.cancelling = False

        # bytes transferred on a current transfer
        self.transferred = 0

        self.speed = SpeedCalc(6)

        for file_name in files:
            parsed_file = Gio.File.parse_name(file_name)

            if not parsed_file.query_exists():
                logging.info(f"Skipping non existing file {parsed_file.get_path()}")
                continue

            file_info = parsed_file.query_info("standard::*", Gio.FileQueryInfoFlags.NONE)

            if file_info.get_file_type() == Gio.FileType.DIRECTORY:
                logging.info(f"Skipping directory {parsed_file.get_path()}")
                continue

            self.files.append(parsed_file)
            self.num_files += 1
            self.total_bytes += file_info.get_size()

        if len(self.files) == 0:
            self.emit("result", False)

        try:
            self.client = Client()
            self.manager.connect_signal('session-added', self.on_session_added)
            self.manager.connect_signal('session-removed', self.on_session_removed)
        except GLib.Error as e:
            if 'StartServiceByName' in e.message:
                logging.debug(e.message)
                parent = self.get_toplevel()
                assert isinstance(parent, Gtk.Container)
                d = ErrorDialog(_("obexd not available"), _("Failed to autostart obex service. Make sure the obex "
                                                            "daemon is running"), parent=parent)
                d.run()
                d.destroy()
                self.emit("result", False)
            else:
                # Fail on anything else
                raise

        basename = self.files[-1].get_basename()
        assert basename is not None
        self.l_file.props.label = basename

        self.client.connect('session-failed', self.on_session_failed)

        logging.info(f"Sending to {device['Address']}")
        self.l_dest.props.label = device['Alias']

        # Stop discovery if discovering and let adapter settle for a second
        if self.adapter["Discovering"]:
            self.adapter.stop_discovery()
            time.sleep(1)

        self.create_session()

        self.show()

    def create_session(self) -> None:
        self.client.create_session(self.device['Address'], self.adapter["Address"])

    def on_cancel(self, button: Optional[Gtk.Button]) -> None:
        self.pb.props.text = _("Cancelling")
        if button:
            button.props.sensitive = False

        if self.object_push:
            self.client.remove_session(self.object_push.get_session_path())

        self.emit("result", False)

    def on_transfer_started(self, _object_push: ObjectPush, transfer_path: str, filename: str) -> None:
        if self.total_transferred == 0:
            self.pb.props.text = _("Sending File") + (" %(0)s/%(1)s (%(2).2f %(3)s/s) " + _("ETA:") + " %(4)s") % {
                "1": self.num_files,
                "0": (self.num_files - len(self.files) + 1),
                "2": 0.0,
                "3": "B/s",
                "4": "∞"}

        self.l_file.props.label = filename
        self._last_bytes = 0
        self.transferred = 0

        self.transfer = Transfer(obj_path=transfer_path)
        self.transfer.connect("error", self.on_transfer_error)
        self.transfer.connect("progress", self.on_transfer_progress)
        self.transfer.connect("completed", self.on_transfer_completed)

    def on_transfer_failed(self, _object_push: ObjectPush, error: str) -> None:
        self.on_transfer_error(None, str(error))

    def on_transfer_progress(self, _transfer: Transfer, progress: int) -> None:
        self.transferred = progress
        if self._last_bytes == 0:
            self.total_transferred += progress
        else:
            self.total_transferred += (progress - self._last_bytes)

        self._last_bytes = progress

        tm = time.time()
        if tm - self._last_update > 0.5:
            spd = self.speed.calc(self.total_transferred)
            (size, units) = format_bytes(spd)
            try:
                x = ((self.total_bytes - self.total_transferred) / spd) + 1
                if x > 60:
                    x /= 60
                    eta = ngettext("%(minutes)d Minute", "%(minutes)d Minutes", round(x)) % {"minutes": round(x)}
                else:
                    eta = ngettext("%(seconds)d Second", "%(seconds)d Seconds", round(x)) % {"seconds": round(x)}
            except ZeroDivisionError:
                eta = "∞"

            self.pb.props.text = _("Sending File") + (" %(0)s/%(1)s (%(2).2f %(3)s/s) " + _("ETA:") + " %(4)s") % {
                "1": self.num_files,
                "0": (self.num_files - len(self.files) + 1),
                "2": size,
                "3": units,
                "4": eta}
            self._last_update = tm

        self.pb.props.fraction = float(self.total_transferred) / self.total_bytes

    def on_transfer_completed(self, _transfer: Transfer) -> None:
        del self.files[-1]
        self.transfer = None

        self.process_queue()

    def process_queue(self) -> None:
        if len(self.files) > 0:
            path = self.files[-1].get_path()
            assert path is not None
            self.send_file(path)
        else:
            self.emit("result", True)

    def send_file(self, file_path: str) -> None:
        logging.info(file_path)
        if self.object_push:
            self.object_push.send_file(file_path)

    def on_transfer_error(self, _transfer: Optional[Transfer], msg: str = "") -> None:
        if not self.error_dialog:
            self.speed.reset()
            parent = self.get_toplevel()
            assert isinstance(parent, Gtk.Container)
            d = ErrorDialog(msg, _("Error occurred while sending file %s") % self.files[-1].get_basename(),
                            modal=True, icon_name="blueman", parent=parent, buttons=Gtk.ButtonsType.NONE)

            if len(self.files) > 1:
                d.add_button(_("Skip"), Gtk.ResponseType.NO)
            d.add_button(_("Retry"), Gtk.ResponseType.YES)
            d.add_button("_Cancel", Gtk.ResponseType.CANCEL)

            if self.object_push:
                self.client.remove_session(self.object_push.get_object_path())

            def on_response(dialog: Gtk.Dialog, resp: int) -> None:
                dialog.destroy()
                self.error_dialog = None

                if resp == Gtk.ResponseType.CANCEL:
                    self.on_cancel(None)
                elif resp == Gtk.ResponseType.NO:
                    finfo = self.files[-1].query_info('standard::*', Gio.FileQueryInfoFlags.NONE)
                    self.total_bytes -= finfo.get_size()
                    self.total_transferred -= self.transferred
                    self.transferred = 0
                    del self.files[-1]
                    if not self.object_push:
                        self.create_session()
                    self.process_queue()
                elif resp == Gtk.ResponseType.YES:
                    self.total_transferred -= self.transferred
                    self.transferred = 0
                    if not self.object_push:
                        self.create_session()

                    self.process_queue()
                else:
                    self.on_cancel(None)

            d.connect("response", on_response)
            d.show()
            self.error_dialog = d

    def on_session_added(self, _manager: Manager, session_path: str) -> None:
        self.object_push = ObjectPush(obj_path=session_path)
        self.object_push_handlers.append(self.object_push.connect("transfer-started", self.on_transfer_started))
        self.object_push_handlers.append(self.object_push.connect("transfer-failed", self.on_transfer_failed))
        self.process_queue()

    def on_session_removed(self, _manager: Manager, session_path: str) -> None:
        logging.debug(f"Session removed: {session_path}")
        if self.object_push:
            for handlerid in self.object_push_handlers:
                self.object_push.disconnect(handlerid)
            self.object_push = None

    def on_session_failed(self, _client: Client, msg: BluezDBusException) -> None:
        parent = self.get_toplevel()
        assert isinstance(parent, Gtk.Container)
        d = ErrorDialog(_("Error occurred"), msg.reason.split(None, 1)[1], icon_name="blueman", parent=parent)

        d.run()
        d.destroy()
        self.emit("result", False)



from gettext import gettext as _
from typing import Optional

from gi.repository import GObject, GLib

from blueman.Functions import launch
from blueman.main.PluginManager import PluginManager
from blueman.plugins.AppletPlugin import AppletPlugin
from blueman.bluemantyping import GSignals


class StatusIconImplementationProvider:
    def on_query_status_icon_implementation(self) -> Optional[str]:
        ...


class StatusIconVisibilityHandler:
    def on_query_force_status_icon_visibility(self) -> bool:
        ...


class StatusIconProvider:
    def on_status_icon_query_icon(self) -> Optional[str]:
        ...


class StatusIcon(AppletPlugin, GObject.GObject):
    __gsignals__: GSignals = {'activate': (GObject.SignalFlags.NO_HOOKS, None, ())}

    __unloadable__ = False
    __icon__ = "blueman-tray"
    __depends__ = ['Menu']

    visible = None

    visibility_timeout = None

    _implementation = None

    def on_load(self) -> None:
        GObject.GObject.__init__(self)
        self.lines = {0: _("Bluetooth Enabled")}

        self.query_visibility(emit=False)

        self.parent.Plugins.connect('plugin-loaded', self._on_plugins_changed)
        self.parent.Plugins.connect('plugin-unloaded', self._on_plugins_changed)

        self._add_dbus_method("GetVisibility", (), "b", lambda: self.visible)
        self._add_dbus_signal("VisibilityChanged", "b")
        self._add_dbus_signal("TextChanged", "s")
        self._add_dbus_method("GetText", (), "s", self._get_text)
        self._add_dbus_signal("IconNameChanged", "s")
        self._add_dbus_method("GetStatusIconImplementation", (), "s", self._get_status_icon_implementation)
        self._add_dbus_method("GetIconName", (), "s", self._get_icon_name)
        self._add_dbus_method("Activate", (), "", lambda: self.emit("activate"))

    def query_visibility(self, delay_hiding: bool = False, emit: bool = True) -> None:
        if self.parent.Manager.get_adapters() or \
           any(plugin.on_query_force_status_icon_visibility()
               for plugin in self.parent.Plugins.get_loaded_plugins(StatusIconVisibilityHandler)):
            self.set_visible(True, emit)
        elif not self.visibility_timeout:
            if delay_hiding:
                self.visibility_timeout = GLib.timeout_add(1000, self.on_visibility_timeout)
            else:
                self.set_visible(False, emit)

    def on_visibility_timeout(self) -> bool:
        assert self.visibility_timeout is not None
        GLib.source_remove(self.visibility_timeout)
        self.visibility_timeout = None
        self.query_visibility()
        return False

    def set_visible(self, visible: bool, emit: bool) -> None:
        self.visible = visible
        if emit:
            self._emit_dbus_signal("VisibilityChanged", visible)

    def set_text_line(self, lineid: int, text: str) -> None:
        if text:
            self.lines[lineid] = text
        else:
            self.lines.pop(lineid, None)

        self._emit_dbus_signal("TextChanged", self._get_text())

    def _get_text(self) -> str:
        return '\n'.join([self.lines[key] for key in sorted(self.lines)])

    def icon_should_change(self) -> None:
        self._emit_dbus_signal("IconNameChanged", self._get_icon_name())
        self.query_visibility()

    def on_adapter_added(self, _path: str) -> None:
        self.query_visibility()

    def on_adapter_removed(self, _path: str) -> None:
        self.query_visibility()

    def on_manager_state_changed(self, state: bool) -> None:
        self.query_visibility()
        if state:
            launch('blueman-tray', icon_name='blueman', sn=False)

    def _on_plugins_changed(self, _plugins: PluginManager, _name: str) -> None:
        implementation = self._get_status_icon_implementation()
        if not self._implementation or self._implementation != implementation:
            self._implementation = implementation

        if self.parent.manager_state:
            launch('blueman-tray', icon_name='blueman', sn=False)

    def _get_status_icon_implementation(self) -> str:
        for plugin in self.parent.Plugins.get_loaded_plugins(StatusIconImplementationProvider):
            implementation = plugin.on_query_status_icon_implementation()
            if implementation:
                return implementation
        return "GtkStatusIcon"

    def _get_icon_name(self) -> str:
        for plugin in self.parent.Plugins.get_loaded_plugins(StatusIconProvider):
            icon = plugin.on_status_icon_query_icon()
            if icon is not None:
                return icon
        return "blueman-tray"



from blueman.services.meta import NetworkService
from blueman.Sdp import GN_SVCLASS_ID


class GroupNetwork(NetworkService):
    __svclass_id__ = GN_SVCLASS_ID
    __icon__ = "network-wireless"
    __priority__ = 80



#!/usr/bin/env python3
# coding=utf-8
#cython: language_level=3
import logging

cdef extern from "malloc.h":
    cdef void free(void *ptr)

cdef extern from "string.h":
    cdef char* strerror(int errnum)

cdef extern from "bluetooth/bluetooth.h":
    ctypedef struct bdaddr_t:
        unsigned char b[6]

    int ba2str(bdaddr_t *ba, char *str)
    int str2ba(char *str, bdaddr_t *ba)

cdef extern from "bluetooth/hci.h":
    cdef struct hci_dev_stats:
        unsigned long int err_rx
        unsigned long int err_tx
        unsigned long int cmd_tx
        unsigned long int evt_rx
        unsigned long int acl_tx
        unsigned long int acl_rx
        unsigned long int sco_tx
        unsigned long int sco_rx
        unsigned long int byte_rx
        unsigned long int byte_tx


    cdef struct hci_dev_info:
        unsigned int dev_id
        char name[8]

        bdaddr_t bdaddr

        unsigned long int flags
        unsigned char type

        unsigned char features[8]

        unsigned long int pkt_type
        unsigned long int link_policy
        unsigned long int link_mode

        unsigned int acl_mtu
        unsigned int acl_pkts
        unsigned int sco_mtu
        unsigned int sco_pkts

        hci_dev_stats stat



cdef extern from "bluetooth/hci_lib.h":
    cdef int hci_devinfo(int dev_id, hci_dev_info *di)


cdef extern from "bluetooth/rfcomm.h":

    cdef struct rfcomm_dev_info:
        short id
        unsigned int flags
        unsigned short state
        bdaddr_t src
        bdaddr_t dst
        unsigned char channel

    cdef struct rfcomm_dev_list_req:
        unsigned short dev_num
        rfcomm_dev_info dev_info[0]


cdef extern from "libblueman.h":
    cdef struct conn_info_handles:
        unsigned int handle
        int dd


    cdef int connection_init(int dev_id, char *addr, conn_info_handles *ci)
    cdef int connection_get_rssi(conn_info_handles *ci, signed char *ret_rssi)
    cdef int connection_get_lq(conn_info_handles *ci, unsigned char *ret_lq)
    cdef int connection_get_tpl(conn_info_handles *ci, signed char *ret_tpl, unsigned char type)
    cdef int connection_close(conn_info_handles *ci)
    cdef int c_get_rfcomm_channel "get_rfcomm_channel" (unsigned short service_class, char* btd_addr)
    cdef int get_rfcomm_list(rfcomm_dev_list_req **ret)
    cdef int c_create_rfcomm_device "create_rfcomm_device" (char *local_address, char *remote_address, int channel)
    cdef int c_release_rfcomm_device "release_rfcomm_device" (int id)
    cdef float get_page_timeout(int hdev)
    cdef int _create_bridge(char* name)
    cdef int _destroy_bridge(char* name)


class RFCOMMError(Exception):
    pass

ERR = {
    -1:"Can't allocate memory",
    -2:"HCI device open failed",
    -3:"Not connected",
    -4:"Get connection info failed",
    -5:"Read RSSI failed",
    -6:"Read transmit power level request failed",
    -7:"Read Link quality failed",
    -8:"Getting rfcomm list failed",
    -9:"ERR_SOCKET_FAILED",
    -10:"ERR_CANT_READ_PAGE_TIMEOUT",
    -11:"ERR_READ_PAGE_TIMEOUT",
    -12: "Can't bind RFCOMM socket",
    -13: "Can't connect RFCOMM socket",
    -14: "Can't create RFCOMM TTY",
    -15: "Can't release RFCOMM TTY"
    }

RFCOMM_STATES = [
    "unknown",
    "connected",
    "clean",
    "bound",
    "listening",
    "connecting",
    "connecting",
    "config",
    "disconnecting",
    "closed"
]
RFCOMM_REUSE_DLC =    0
RFCOMM_RELEASE_ONHUP =    1
RFCOMM_HANGUP_NOW =    2
RFCOMM_TTY_ATTACHED =    3

def get_rfcomm_channel(uuid, py_bdaddr):
    if py_bdaddr is None:
        return

    py_bytes = py_bdaddr.encode('UTF-8')
    cdef char* bdaddr = py_bytes
    return c_get_rfcomm_channel(uuid, bdaddr)

def rfcomm_list():
    cdef rfcomm_dev_list_req *dl

    cdef char src[18]
    cdef char dst[18]


    res = get_rfcomm_list(&dl)
    if res < 0:
        raise Exception, ERR[res]

    devs = []
    for 0 <= i < dl.dev_num:
        ba2str(&dl.dev_info[i].src, src)
        ba2str(&dl.dev_info[i].dst, dst)

        devs.append(
            {    "id": dl.dev_info[i].id,
                "channel": dl.dev_info[i].channel,
                "flags": dl.dev_info[i].flags,
                "state": RFCOMM_STATES[dl.dev_info[i].state],
                "src": src.decode("UTF-8"),
                "dst": dst.decode("UTF-8")
            })

    free(dl)

    return devs


def create_rfcomm_device(py_local_address, py_remote_address, channel):
    py_bytes_local_address = py_local_address.encode('UTF-8')
    py_bytes_remote_address = py_remote_address.encode('UTF-8')
    cdef char* local_address = py_bytes_local_address
    cdef char* remote_address = py_bytes_remote_address
    ret = c_create_rfcomm_device(local_address, remote_address, channel)
    if ret < 0:
        raise RFCOMMError(ERR[ret])
    return ret


def release_rfcomm_device(id):
    ret = c_release_rfcomm_device(id)
    if ret < 0:
        raise RFCOMMError(ERR[ret])
    return ret


class BridgeException(Exception):
    def __init__(self, errno):
        self.errno = errno

    def __str__(self):
        return strerror(self.errno)



def create_bridge(py_name="pan1"):
    py_bytes_name = py_name.encode("UTF-8")
    cdef char* name = py_bytes_name

    err = _create_bridge(name)
    if err < 0:
        raise BridgeException(-err)

def destroy_bridge(py_name="pan1"):
    py_bytes_name = py_name.encode("UTF-8")
    cdef char* name = py_bytes_name

    err = _destroy_bridge(name)
    if err < 0:
        raise BridgeException(-err)


class ConnInfoReadError(Exception):
    pass

cdef class conn_info:
    cdef conn_info_handles ci
    cdef int hci
    cdef char* addr
    cdef public bint failed

    def __init__(self, py_addr, py_hci_name="hci0"):
        self.failed = False
        py_bytes_addr = py_addr.encode("UTF-8")
        cdef char* addr = py_bytes_addr

        py_bytes_hci_name = py_hci_name.encode("UTF-8")
        cdef char* hci_name = py_bytes_hci_name


        self.hci = int(hci_name[3:])
        self.addr = addr

    def init(self):
        res = connection_init(self.hci, self.addr, & self.ci)
        if res < 0:
            self.failed = True
            raise ConnInfoReadError(ERR[res])

    def deinit(self):
        if self.failed:
            return
        connection_close(&self.ci)

    def get_rssi(self):
        cdef signed char rssi
        res = connection_get_rssi(&self.ci, &rssi)
        if res < 0:
            raise ConnInfoReadError(ERR[res])

        return rssi

    def get_lq(self):
        cdef unsigned char lq
        res = connection_get_lq(&self.ci, &lq)
        if res < 0:
            raise ConnInfoReadError(ERR[res])

        return lq

    def get_tpl(self, tp=0):
        cdef signed char tpl
        res = connection_get_tpl(&self.ci, &tpl, tp)
        if res < 0:
            raise ConnInfoReadError(ERR[res])

        return tpl

def page_timeout(py_hci_name="hci0"):
    py_bytes_hci_name = py_hci_name.encode("UTF-8")
    cdef char* hci_name = py_bytes_hci_name

    dev_id = int(hci_name[3:])
    ret = get_page_timeout(dev_id)
    if ret < 0:
        raise Exception, ERR[ret]
    else:
        return ret

def device_info(py_hci_name="hci0"):
    py_bytes_hci_name = py_hci_name.encode("UTF-8")
    cdef char* hci_name = py_bytes_hci_name

    cdef hci_dev_info di
    cdef int ret

    dev_id = int(hci_name[3:])

    res = hci_devinfo(dev_id, &di)

    cdef char addr[32]
    ba2str(&di.bdaddr, addr)

    feats = []
    for 0 <= i < 8:
        feats.append(di.features[i])

    x = [("err_rx", di.stat.err_rx),
    ("err_tx",di.stat.err_tx),
    ("cmd_tx",di.stat.cmd_tx),
    ("evt_rx",di.stat.evt_rx),
    ("acl_tx",di.stat.acl_tx),
    ("acl_rx",di.stat.acl_rx),
    ("sco_tx",di.stat.sco_tx),
    ("sco_rx",di.stat.sco_rx),
    ("byte_rx",di.stat.byte_rx),
    ("byte_tx",di.stat.byte_tx)]

    z = [("dev_id", di.dev_id),
    ("name", di.name.decode("UTF-8")),
    ("bdaddr",addr.decode("UTF-8")),
    ("flags",di.flags),
    ("type",di.type),
    ("features",feats),
    ("pkt_type",di.pkt_type),
    ("link_policy",di.link_policy),
    ("link_mode",di.link_mode),
    ("acl_mtu",di.acl_mtu),
    ("acl_pkts",di.acl_pkts),
    ("sco_mtu",di.sco_pkts),
    ("stat", dict(x))]

    return dict(z)

cdef extern from "glib-object.h":
    ctypedef struct GObject

cdef extern from "pygobject.h":
    cdef GObject* pygobject_get(object)



from gettext import gettext as _
from random import randint
import logging
import ipaddress
from typing import List, Tuple, cast, Union, TYPE_CHECKING

from blueman.Functions import have, get_local_interfaces
from blueman.main.Builder import Builder
from blueman.plugins.ServicePlugin import ServicePlugin
from blueman.main.NetConf import NetConf, DnsMasqHandler, DhcpdHandler, UdhcpdHandler
from blueman.main.Config import Config
from blueman.main.DBusProxies import Mechanism
from blueman.main.DBusProxies import AppletService
from blueman.gui.CommonUi import ErrorDialog

import gi
gi.require_version("Gtk", "3.0")
from gi.repository import Gio, Gtk, GObject

if TYPE_CHECKING:
    from typing_extensions import Literal


class Network(ServicePlugin):
    __plugin_info__ = (_("Network"), "network-workgroup")

    def on_load(self, container: Gtk.Box) -> None:

        self._builder = Builder("services-network.ui")
        self.widget = self._builder.get_widget("network_frame", Gtk.Widget)

        container.pack_start(self.widget, True, True, 0)

        self.interfaces: List[Tuple[str, ipaddress.IPv4Interface]] = []
        netifs = get_local_interfaces()
        for iface in netifs:
            if iface != "lo" and iface != "pan1":
                logging.info(iface)
                ipiface = ipaddress.ip_interface('/'.join(cast(Tuple[str, str], netifs[iface])))
                self.interfaces.append((iface, ipiface))

        self.setup_network()
        try:
            self.ip_check()
        except (ValueError, ipaddress.AddressValueError) as e:
            logging.exception(e)

    def on_enter(self) -> None:
        self.widget.props.visible = True

    def on_leave(self) -> None:
        self.widget.props.visible = False

    def on_apply(self) -> None:

        if self.on_query_apply_state():
            logging.info("network apply")

            m = Mechanism()
            nap_enable = self._builder.get_widget("nap-enable", Gtk.CheckButton)
            if nap_enable.props.active:

                if self._builder.get_widget("r_dhcpd", Gtk.RadioButton).props.active:
                    stype = "DhcpdHandler"
                elif self._builder.get_widget("r_dnsmasq", Gtk.RadioButton).props.active:
                    stype = "DnsMasqHandler"
                elif self._builder.get_widget("r_udhcpd", Gtk.RadioButton).props.active:
                    stype = "UdhcpdHandler"

                net_ip = self._builder.get_widget("net_ip", Gtk.Entry)

                try:
                    m.EnableNetwork('(sss)', net_ip.props.text, "255.255.255.0", stype)

                    if not self.Config["nap-enable"]:
                        self.Config["nap-enable"] = True
                except Exception as e:
                    parent = self.widget.get_toplevel()
                    assert isinstance(parent, Gtk.Container)
                    d = ErrorDialog("<b>Failed to apply network settings</b>", excp=e, parent=parent)

                    d.run()
                    d.destroy()
                    return
            else:
                self.Config["nap-enable"] = False
                m.DisableNetwork()

            self.clear_options()

    def ip_check(self) -> None:
        entry = self._builder.get_widget("net_ip", Gtk.Entry)
        try:
            nap_ipiface = ipaddress.ip_interface('/'.join((entry.props.text, '255.255.255.0')))
        except (ValueError, ipaddress.AddressValueError):
            entry.props.secondary_icon_name = "dialog-error"
            entry.props.secondary_icon_tooltip_text = _("Invalid IP address")
            raise

        for iface, ipiface in self.interfaces:
            if nap_ipiface.ip == ipiface.ip:
                error_message = _("IP address conflicts with interface %s which has the same address" % iface)
                tooltip_text = error_message
                entry.props.secondary_icon_name = "dialog-error"
                entry.props.secondary_icon_tooltip_text = tooltip_text
                raise ValueError(error_message)

            elif nap_ipiface.network == ipiface.network:
                tooltip_text = _(
                    "IP address overlaps with subnet of interface %s, which has the following configuration  %s/%s\n"
                    "This may cause incorrect network behavior" % (iface, ipiface.ip, ipiface.netmask))
                entry.props.secondary_icon_name = "dialog-warning"
                entry.props.secondary_icon_tooltip_text = tooltip_text
                return

        entry.props.secondary_icon_name = None

    def on_query_apply_state(self) -> Union[bool, "Literal[-1]"]:
        opts = self.get_options()
        if not opts:
            return False

        if "ip" in opts:
            try:
                self.ip_check()
            except (ValueError, ipaddress.AddressValueError) as e:
                logging.exception(e)
                return -1

        return True

    def setup_network(self) -> None:
        self.Config = Config("org.blueman.network")

        nap_enable = self._builder.get_widget("nap-enable", Gtk.CheckButton)
        r_dnsmasq = self._builder.get_widget("r_dnsmasq", Gtk.RadioButton)
        r_dhcpd = self._builder.get_widget("r_dhcpd", Gtk.RadioButton)
        r_udhcpd = self._builder.get_widget("r_udhcpd", Gtk.RadioButton)
        net_ip = self._builder.get_widget("net_ip", Gtk.Entry)
        rb_nm = self._builder.get_widget("rb_nm", Gtk.RadioButton)
        rb_blueman = self._builder.get_widget("rb_blueman", Gtk.RadioButton)
        rb_dun_nm = self._builder.get_widget("rb_dun_nm", Gtk.RadioButton)
        rb_dun_blueman = self._builder.get_widget("rb_dun_blueman", Gtk.RadioButton)

        nap_frame = self._builder.get_widget("nap_frame", Gtk.Frame)
        warning = self._builder.get_widget("warning", Gtk.Box)

        if not self.Config["nap-enable"]:
            nap_frame.props.sensitive = False

        nc = NetConf.get_default()
        if nc.ip4_address is not None:
            # previously we stored a bytearray, ipaddress module reads both
            net_ip.props.text = str(ipaddress.ip_address(nc.ip4_address))
            nap_enable.props.active = True
        else:
            net_ip.props.text = "10.%d.%d.1" % (randint(0, 255), randint(0, 255))

        if nc.get_dhcp_handler() is None:
            nap_frame.props.sensitive = False
            nap_enable.props.active = False
            r_dnsmasq.props.active = True
            self.Config["nap-enable"] = False

        have_dhcpd = have("dhcpd3") or have("dhcpd")
        have_dnsmasq = have("dnsmasq")
        have_udhcpd = have("udhcpd")

        if nc.get_dhcp_handler() == DnsMasqHandler and have_dnsmasq:
            r_dnsmasq.props.active = True
        elif nc.get_dhcp_handler() == DhcpdHandler and have_dhcpd:
            r_dhcpd.props.active = True
        elif nc.get_dhcp_handler() == UdhcpdHandler and have_udhcpd:
            r_udhcpd.props.active = True
        else:
            r_dnsmasq.props.active = True

        if not have_dnsmasq and not have_dhcpd and not have_udhcpd:
            nap_frame.props.sensitive = False
            warning.props.visible = True
            warning.props.sensitive = True
            nap_enable.props.sensitive = False
            self.Config["nap-enable"] = False

        if not have_dnsmasq:
            r_dnsmasq.props.sensitive = False
            r_dnsmasq.props.active = False

        if not have_dhcpd:
            r_dhcpd.props.sensitive = False
            r_dhcpd.props.active = False

        if not have_udhcpd:
            r_udhcpd.props.sensitive = False
            r_udhcpd.props.active = False

        r_dnsmasq.connect("toggled", lambda x: self.option_changed_notify("dnsmasq"))
        r_dhcpd.connect("toggled", lambda x: self.option_changed_notify("dhcpd"))
        r_udhcpd.connect("toggled", lambda x: self.option_changed_notify("udhcpd"))

        net_ip.connect("changed", lambda x: self.option_changed_notify("ip", False))
        nap_enable.connect("toggled", lambda x: self.option_changed_notify("nap_enable"))

        self.Config.bind_to_widget("nap-enable", nap_enable, "active", Gio.SettingsBindFlags.GET)

        nap_enable.bind_property("active", nap_frame, "sensitive", GObject.BindingFlags.DEFAULT)

        applet = AppletService()

        avail_plugins = applet.QueryAvailablePlugins()
        active_plugins = applet.QueryPlugins()

        def dun_support_toggled(rb: Gtk.RadioButton, x: str) -> None:
            if rb.props.active and x == "nm":
                applet.SetPluginConfig('(sb)', "PPPSupport", False)
                applet.SetPluginConfig('(sb)', "NMDUNSupport", True)
            elif rb.props.active and x == "blueman":
                applet.SetPluginConfig('(sb)', "NMDUNSupport", False)
                applet.SetPluginConfig('(sb)', "PPPSupport", True)

        def pan_support_toggled(rb: Gtk.RadioButton, x: str) -> None:
            if rb.props.active and x == "nm":
                applet.SetPluginConfig('(sb)', "DhcpClient", False)
                applet.SetPluginConfig('(sb)', "NMPANSupport", True)

            elif rb.props.active and x == "blueman":
                applet.SetPluginConfig('(sb)', "NMPANSupport", False)
                applet.SetPluginConfig('(sb)', "DhcpClient", True)

        if "PPPSupport" in active_plugins:
            rb_dun_blueman.props.active = True

        if "NMDUNSupport" in avail_plugins:
            rb_dun_nm.props.sensitive = True
        else:
            rb_dun_nm.props.sensitive = False
            rb_dun_nm.props.tooltip_text = _("Not currently supported with this setup")

        if "DhcpClient" in active_plugins:
            rb_blueman.props.active = True

        if "NMPANSupport" in avail_plugins:
            rb_nm.props.sensitive = True
        else:
            rb_nm.props.sensitive = False
            rb_nm.props.tooltip_text = _("Not currently supported with this setup")

        if "NMPANSupport" in active_plugins:
            rb_nm.props.active = True

        if "NMDUNSupport" in active_plugins:
            rb_dun_nm.props.active = True

        rb_nm.connect("toggled", pan_support_toggled, "nm")
        rb_blueman.connect("toggled", pan_support_toggled, "blueman")

        rb_dun_nm.connect("toggled", dun_support_toggled, "nm")
        rb_dun_blueman.connect("toggled", dun_support_toggled, "blueman")



from gi.repository import GLib


class BluezDBusException(Exception):
    def __init__(self, reason: str):
        self.reason = reason

    def __str__(self) -> str:
        return self.reason


class DBusFailedError(BluezDBusException):
    pass


class DBusInvalidArgumentsError(BluezDBusException):
    pass


class DBusNotAuthorizedError(BluezDBusException):
    pass


class DBusOutOfMemoryError(BluezDBusException):
    pass


class DBusNoSuchAdapterError(BluezDBusException):
    pass


class DBusNotReadyError(BluezDBusException):
    pass


class DBusNotAvailableError(BluezDBusException):
    pass


class DBusNotConnectedError(BluezDBusException):
    pass


class DBusConnectionAttemptFailedError(BluezDBusException):
    pass


class DBusAlreadyExistsError(BluezDBusException):
    pass


class DBusDoesNotExistError(BluezDBusException):
    pass


class DBusNoReplyError(BluezDBusException):
    pass


class DBusInProgressError(BluezDBusException):
    pass


class DBusNotSupportedError(BluezDBusException):
    pass


class DBusAuthenticationFailedError(BluezDBusException):
    pass


class DBusAuthenticationTimeoutError(BluezDBusException):
    pass


class DBusAuthenticationRejectedError(BluezDBusException):
    pass


class DBusAuthenticationCanceledError(BluezDBusException):
    pass


class DBusUnsupportedMajorClassError(BluezDBusException):
    pass


class DBusServiceUnknownError(BluezDBusException):
    pass


class DBusMainLoopNotSupportedError(BluezDBusException):
    pass


class DBusMainLoopModuleNotFoundError(BluezDBusException):
    pass


class BluezUnavailableAgentMethodError(BluezDBusException):
    pass


__DICT_ERROR__ = {'org.bluez.Error.Failed': DBusFailedError,
                  'org.bluez.Error.InvalidArguments': DBusInvalidArgumentsError,
                  'org.bluez.Error.NotAuthorized': DBusNotAuthorizedError,
                  'org.bluez.Error.OutOfMemory': DBusOutOfMemoryError,
                  'org.bluez.Error.NoSuchAdapter': DBusNoSuchAdapterError,
                  'org.bluez.Error.NotReady': DBusNotReadyError,
                  'org.bluez.Error.NotAvailable': DBusNotAvailableError,
                  'org.bluez.Error.NotConnected': DBusNotConnectedError,
                  'org.bluez.serial.Error.ConnectionAttemptFailed': DBusConnectionAttemptFailedError,
                  'org.bluez.Error.AlreadyExists': DBusAlreadyExistsError,
                  'org.bluez.Error.DoesNotExist': DBusDoesNotExistError,
                  'org.bluez.Error.InProgress': DBusInProgressError,
                  'org.bluez.Error.NoReply': DBusNoReplyError,
                  'org.bluez.Error.NotSupported': DBusNotSupportedError,
                  'org.bluez.Error.AuthenticationFailed': DBusAuthenticationFailedError,
                  'org.bluez.Error.AuthenticationTimeout': DBusAuthenticationTimeoutError,
                  'org.bluez.Error.AuthenticationRejected': DBusAuthenticationRejectedError,
                  'org.bluez.Error.AuthenticationCanceled': DBusAuthenticationCanceledError,
                  'org.bluez.serial.Error.NotSupported': DBusNotSupportedError,
                  'org.bluez.Error.UnsupportedMajorClass': DBusUnsupportedMajorClassError,
                  'org.freedesktop.DBus.Error.ServiceUnknown': DBusServiceUnknownError}


def parse_dbus_error(exception: GLib.Error) -> BluezDBusException:
    global __DICT_ERROR__

    gerror, dbus_error, message = exception.message.split(':', 2)
    try:
        return __DICT_ERROR__[dbus_error](message)
    except KeyError:
        return BluezDBusException(dbus_error + message)



import builtins
import typing
import cairo

from gi.repository import Gdk
from gi.repository import GdkPixbuf
from gi.repository import Gio
from gi.repository import Pango
from gi.repository import xlib


class X11AppLaunchContext(Gdk.AppLaunchContext):
    ...


class X11Cursor(Gdk.Cursor):

    def get_xcursor(self) -> builtins.int: ...

    def get_xdisplay(self) -> xlib.Display: ...


class X11DeviceCore(Gdk.Device):
    ...


class X11DeviceManagerCore(Gdk.DeviceManager):
    ...


class X11DeviceXI2(Gdk.Device):
    ...


class X11Display(Gdk.Display):

    def error_trap_pop(self) -> builtins.int: ...

    def error_trap_pop_ignored(self) -> None: ...

    def error_trap_push(self) -> None: ...

    @staticmethod
    def get_glx_version(display: Gdk.Display) -> typing.Tuple[builtins.bool, builtins.int, builtins.int]: ...

    def get_startup_notification_id(self) -> builtins.str: ...

    def get_user_time(self) -> builtins.int: ...

    def get_xdisplay(self) -> xlib.Display: ...

    def grab(self) -> None: ...

    def set_cursor_theme(self, theme: typing.Optional[builtins.str], size: builtins.int) -> None: ...

    def set_startup_notification_id(self, startup_id: builtins.str) -> None: ...

    def set_window_scale(self, scale: builtins.int) -> None: ...

    def string_to_compound_text(self, str: builtins.str) -> typing.Tuple[builtins.int, Gdk.Atom, builtins.int, builtins.bytes]: ...

    def text_property_to_text_list(self, encoding: Gdk.Atom, format: builtins.int, text: builtins.int, length: builtins.int, list: builtins.str) -> builtins.int: ...

    def ungrab(self) -> None: ...

    def utf8_to_compound_text(self, str: builtins.str) -> typing.Tuple[builtins.bool, Gdk.Atom, builtins.int, builtins.bytes]: ...


class X11DisplayManager(Gdk.DisplayManager):
    ...


class X11DragContext(Gdk.DragContext):
    ...


class X11GLContext(Gdk.GLContext):
    ...


class X11Keymap(Gdk.Keymap):

    def get_group_for_state(self, state: builtins.int) -> builtins.int: ...

    def key_is_modifier(self, keycode: builtins.int) -> builtins.bool: ...


class X11Monitor(Gdk.Monitor):

    @staticmethod
    def get_output(monitor: Gdk.Monitor) -> builtins.int: ...


class X11Screen(Gdk.Screen):

    def get_current_desktop(self) -> builtins.int: ...

    def get_monitor_output(self, monitor_num: builtins.int) -> builtins.int: ...

    def get_number_of_desktops(self) -> builtins.int: ...

    def get_screen_number(self) -> builtins.int: ...

    def get_window_manager_name(self) -> builtins.str: ...

    def get_xscreen(self) -> xlib.Screen: ...

    def lookup_visual(self, xvisualid: builtins.int) -> X11Visual: ...

    def supports_net_wm_hint(self, property: Gdk.Atom) -> builtins.bool: ...


class X11Visual(Gdk.Visual):

    def get_xvisual(self) -> xlib.Visual: ...


class X11Window(Gdk.Window):

    @staticmethod
    def foreign_new_for_display(display: X11Display, window: builtins.int) -> Gdk.Window: ...

    def get_desktop(self) -> builtins.int: ...

    def get_xid(self) -> builtins.int: ...

    @staticmethod
    def lookup_for_display(display: X11Display, window: builtins.int) -> X11Window: ...

    def move_to_current_desktop(self) -> None: ...

    def move_to_desktop(self, desktop: builtins.int) -> None: ...

    def set_frame_extents(self, left: builtins.int, right: builtins.int, top: builtins.int, bottom: builtins.int) -> None: ...

    def set_frame_sync_enabled(self, frame_sync_enabled: builtins.bool) -> None: ...

    def set_hide_titlebar_when_maximized(self, hide_titlebar_when_maximized: builtins.bool) -> None: ...

    def set_theme_variant(self, variant: builtins.str) -> None: ...

    def set_user_time(self, timestamp: builtins.int) -> None: ...

    def set_utf8_property(self, name: builtins.str, value: typing.Optional[builtins.str]) -> None: ...


class X11DeviceManagerXI2(X11DeviceManagerCore):
    ...


def x11_atom_to_xatom(atom: Gdk.Atom) -> builtins.int: ...


def x11_atom_to_xatom_for_display(display: X11Display, atom: Gdk.Atom) -> builtins.int: ...


def x11_device_get_id(device: X11DeviceCore) -> builtins.int: ...


def x11_device_manager_lookup(device_manager: X11DeviceManagerCore, device_id: builtins.int) -> typing.Optional[X11DeviceCore]: ...


def x11_free_compound_text(ctext: builtins.int) -> None: ...


def x11_free_text_list(list: builtins.str) -> None: ...


def x11_get_default_root_xwindow() -> builtins.int: ...


def x11_get_default_screen() -> builtins.int: ...


def x11_get_default_xdisplay() -> xlib.Display: ...


def x11_get_parent_relative_pattern() -> cairo.Pattern: ...


def x11_get_server_time(window: X11Window) -> builtins.int: ...


def x11_get_xatom_by_name(atom_name: builtins.str) -> builtins.int: ...


def x11_get_xatom_by_name_for_display(display: X11Display, atom_name: builtins.str) -> builtins.int: ...


def x11_get_xatom_name(xatom: builtins.int) -> builtins.str: ...


def x11_get_xatom_name_for_display(display: X11Display, xatom: builtins.int) -> builtins.str: ...


def x11_grab_server() -> None: ...


def x11_lookup_xdisplay(xdisplay: xlib.Display) -> X11Display: ...


def x11_register_standard_event_type(display: X11Display, event_base: builtins.int, n_events: builtins.int) -> None: ...


def x11_set_sm_client_id(sm_client_id: typing.Optional[builtins.str]) -> None: ...


def x11_ungrab_server() -> None: ...


def x11_xatom_to_atom(xatom: builtins.int) -> Gdk.Atom: ...


def x11_xatom_to_atom_for_display(display: X11Display, xatom: builtins.int) -> Gdk.Atom: ...





from typing import List, Tuple, Union, TYPE_CHECKING


import gi
gi.require_version("Gtk", "3.0")
from gi.repository import Gtk

if TYPE_CHECKING:
    from typing_extensions import Literal

    from blueman.main.Services import BluemanServices


class ServicePlugin:
    _options: List[str]
    instances: List["ServicePlugin"] = []
    __plugin_info__: Tuple[str, str]

    def __init__(self, parent: "BluemanServices"):
        ServicePlugin.instances.append(self)
        self._options = []
        self.parent = parent

        self.__is_exposed = False
        self._is_loaded = False

    def _on_enter(self) -> None:
        if not self.__is_exposed:
            self.on_enter()
            self.__is_exposed = True

    def _on_leave(self) -> None:
        if self.__is_exposed:
            self.on_leave()
            self.__is_exposed = False

    # call when option has changed.
    def option_changed_notify(self, option_id: str, state: bool = True) -> None:

        if option_id not in self._options:
            self._options.append(option_id)
        else:
            if state:
                self._options.remove(option_id)

        self.parent.option_changed()

    def get_options(self) -> List[str]:
        return self._options

    def clear_options(self) -> None:
        self._options = []

    def on_load(self, container: Gtk.Box) -> None:
        pass

    def on_unload(self) -> None:
        pass

    # return true if apply button should be sensitive or false if not. -1 to force disabled
    def on_query_apply_state(self) -> Union[bool, "Literal[-1]"]:
        pass

    def on_apply(self) -> None:
        pass

    # called when current plugin's page is selected. The plugin's widget should be shown
    def on_enter(self) -> None:
        pass

    # called when current plugin's page is changed to another. The plugin's widget should be hidden.
    def on_leave(self) -> None:
        pass



from gettext import gettext as _
import logging
from typing import Any

from blueman.bluez.Device import Device
from blueman.Functions import launch
from blueman.plugins.AppletPlugin import AppletPlugin

import gi
gi.require_version('GdkX11', '3.0')
gi.require_version('Gdk', '3.0')
from gi.repository import Gdk, GdkX11

if not isinstance(Gdk.Screen.get_default(), GdkX11.X11Screen):
    raise ImportError('This is not an X11 screen')


class GameControllerWakelock(AppletPlugin):
    __description__ = _("Temporarily suspends the screensaver when a bluetooth game controller is connected.")
    __author__ = "bwRavencl"
    __icon__ = "input-gaming"

    def on_load(self) -> None:
        self.wake_lock = 0
        screen = Gdk.Screen.get_default()
        assert screen is not None
        window = screen.get_root_window()
        assert isinstance(window, GdkX11.X11Window)
        self.root_window_id = "0x%x" % window.get_xid()

    def on_unload(self) -> None:
        if self.wake_lock:
            self.wake_lock = 1
            self.xdg_screensaver("resume")

    def on_device_property_changed(self, path: str, key: str, value: Any) -> None:
        if key == "Connected":
            klass = Device(obj_path=path)["Class"] & 0x1fff

            if klass == 0x504 or klass == 0x508:
                if value:
                    self.xdg_screensaver("suspend")
                else:
                    self.xdg_screensaver("resume")

    def xdg_screensaver(self, action: str) -> None:
        command = f"xdg-screensaver {action} {self.root_window_id}"

        if action == "resume":
            if self.wake_lock <= 0:
                self.wake_lock = 0
            elif self.wake_lock > 1:
                self.wake_lock -= 1
            else:
                ret = launch(command, sn=False)
                if ret:
                    self.wake_lock -= 1
                else:
                    logging.error(f"{action} failed")

        elif action == "suspend":
            if self.wake_lock >= 1:
                self.wake_lock += 1
            else:
                ret = launch(command, sn=False)
                if ret:
                    self.wake_lock += 1
                else:
                    logging.error(f"{action} failed")

        logging.info(f"Number of locks: {self.wake_lock}")



import gi
import logging
import uuid
from typing import Optional, Callable, Union

from blueman.Service import Service

try:
    gi.require_version('NM', '1.0')
except ValueError:
    raise ImportError('NM python bindings not found.')

from gi.repository import GLib, GObject, NM, Gio
from blueman.main.Config import Config


class NMConnectionError(Exception):
    pass


class NMConnectionBase:
    conntype: str

    def __init__(self, service: Service, reply_handler: Callable[[], None],
                 error_handler: Callable[[Union[NMConnectionError, GLib.Error]], None]):
        if self.conntype not in ('dun', 'panu'):
            error_handler(
                NMConnectionError(f"Invalid connection type {self.conntype}, should be panu or dun")
            )
        self.device = service.device
        self.bdaddr = self.device['Address']
        self.error_handler = error_handler
        self.reply_handler = reply_handler
        self.connection = None
        self.active_connection = None
        self.client = NM.Client.new()
        self.Config = Config("org.blueman.gsmsetting", f"/org/blueman/gsmsettings/{self.bdaddr}/")
        self._statehandler: Optional[int] = None

        self.find_or_create_connection()

    def _on_connection_added(self, client: NM.Client, result: Gio.AsyncResult, conn_uuid: str) -> None:
        try:
            self.connection = client.add_connection_finish(result)
        except GLib.Error as e:
            self.error_handler(e)

        self.store_uuid(conn_uuid)

    def _on_device_state_changed(self, device: NM.Device, new_state: int, old_state: int, reason: int) -> None:
        new = NM.DeviceState(new_state)
        old = NM.DeviceState(old_state)
        state_reason = NM.DeviceStateReason(reason)
        logging.debug(f"New: {new.value_nick} Old: {old.value_nick} Reason: {state_reason.value_nick}")

        error_msg = None

        if new == NM.DeviceState.FAILED:
            error_msg = f"Connection failed with reason: {state_reason.value_nick}"
        elif new == NM.DeviceState.ACTIVATED:
            logging.debug("Connection successfully activated")
        elif (new <= NM.DeviceState.DISCONNECTED or new == NM.DeviceState.DEACTIVATING) and \
                (NM.DeviceState.DISCONNECTED < old <= NM.DeviceState.ACTIVATED):
            error_msg = f"Connection disconnected with reason {state_reason.value_nick}"
        else:
            return  # Keep checking the state changes

        # We are done with state changes
        assert self._statehandler is not None
        GObject.signal_handler_disconnect(device, self._statehandler)
        if error_msg is None:
            self.reply_handler()
        else:
            logging.debug(error_msg)
            self.error_handler(NMConnectionError(error_msg))

    def activate(self) -> None:
        def on_connection_activate(client: NM.Client, result: Gio.AsyncResult) -> None:
            try:
                self.active_connection = client.activate_connection_finish(result)
            except GLib.Error as e:
                self.error_handler(e)

        device = self.client.get_device_by_iface(self.bdaddr)
        if not device:
            self.error_handler(NMConnectionError(f"Could not find device {self.bdaddr}"))
        elif device.get_state() == NM.DeviceState.ACTIVATED:
            self.error_handler(NMConnectionError(f"Device {self.bdaddr} already activated"))
        else:
            self._statehandler = device.connect('state-changed', self._on_device_state_changed)
            self.client.activate_connection_async(self.connection, device, None, None, on_connection_activate)

    def deactivate(self) -> None:
        def on_connection_deactivate(client: NM.Client, result: Gio.AsyncResult) -> None:
            try:
                client.deactivate_connection_finish(result)
                logging.debug(f"Device {self.bdaddr} deactivated sucessfully")
                self.reply_handler()
                self.active_connection = None
            except GLib.Error as e:
                self.error_handler(e)

        self.client.deactivate_connection_async(self.active_connection, None, on_connection_deactivate)

    def find_or_create_connection(self) -> None:
        if not self.connection_uuid:
            self.create_connection()
        else:
            conn = self.client.get_connection_by_uuid(self.connection_uuid)
            if conn is None:
                self.create_connection()
            else:
                logging.debug(f"Found existing connection with uuid {self.connection_uuid}")
                self.connection = conn

                if self.conntype == 'dun':
                    settings_gsm = conn.get_setting_gsm()

                    if settings_gsm.props.apn != self.Config['apn']:
                        logging.debug(f"Updating apn on connection to {self.Config['apn']}")
                        settings_gsm.props.apn = self.Config['apn']
                    if settings_gsm.props.number != self.Config['number']:
                        logging.debug(f"Updating number on connection to {self.Config['number']}")
                        settings_gsm.props.number = self.Config['number']

                conn.commit_changes(True, None)

        # Try to find active connection
        for active_conn in self.client.get_active_connections():
            conn = active_conn.get_connection()
            if conn == self.connection:
                self.active_connection = active_conn

    @property
    def connected(self) -> bool:
        if self.active_connection is None:
            return False

        state = self.active_connection.get_state()
        if state == NM.ActiveConnectionState.CONNECTED:
            return True
        else:
            return False

    def create_connection(self) -> None:
        raise NotImplementedError

    def store_uuid(self, conn_uuid: str) -> None:
        raise NotImplementedError

    @property
    def connection_uuid(self) -> str:
        raise NotImplementedError


class NMPANConnection(NMConnectionBase):
    conntype = 'panu'

    def store_uuid(self, conn_uuid: str) -> None:
        self.Config['nmpanuuid'] = conn_uuid

    @property
    def connection_uuid(self) -> str:
        # PANU connections are automatically created so attempt to find it
        # It appears the Name property is used not Alias!
        conn = self.client.get_connection_by_id(f"{self.device['Name']} Network")
        res: str
        if conn is not None:
            conn_settings = conn.get_setting_connection()
            res = conn_settings.get_uuid()
        else:
            res = self.Config['nmpanuuid']
        return res

    def create_connection(self) -> None:
        conn = NM.SimpleConnection()
        conn_id = f"{self.device['Name']} Network"
        conn_uuid = str(uuid.uuid4())

        conn_sett = NM.SettingConnection(type='bluetooth', id=conn_id, uuid=conn_uuid, autoconnect=False)
        conn_sett_bt = NM.SettingBluetooth(type=self.conntype, bdaddr=self.bdaddr)
        conn.add_setting(conn_sett)
        conn.add_setting(conn_sett_bt)

        self.client.add_connection_async(conn, True, None, self._on_connection_added, conn_uuid)


class NMDUNConnection(NMConnectionBase):
    conntype = 'dun'

    def store_uuid(self, conn_uuid: str) -> None:
        self.Config['nmdunuuid'] = conn_uuid

    @property
    def connection_uuid(self) -> str:
        res: str = self.Config['nmdunuuid']
        return res

    def create_connection(self) -> None:
        if not self.Config['apn']:
            self.error_handler(NMConnectionError('No apn configured, make sure to configure dialup settings'))
            return

        conn = NM.SimpleConnection()
        conn_id = f"blueman dun for {self.device['Alias']}"
        conn_uuid = str(uuid.uuid4())

        conn_sett = NM.SettingConnection(type='bluetooth', id=conn_id, uuid=conn_uuid, autoconnect=False)
        conn_sett_bt = NM.SettingBluetooth(type=self.conntype, bdaddr=self.bdaddr)
        conn_sett_gsm = NM.SettingGsm(apn=self.Config['apn'], number=self.Config['number'])
        conn.add_setting(conn_sett)
        conn.add_setting(conn_sett_bt)
        conn.add_setting(conn_sett_gsm)

        self.client.add_connection_async(conn, True, None, self._on_connection_added, conn_uuid)



from typing import Dict, List, Callable, Optional, Iterable, Tuple, Union, Type

from blueman.main.Config import Config

import gi
gi.require_version("Gtk", "3.0")
gi.require_version("Gdk", "3.0")
from gi.repository import Gtk
from gi.repository import Gdk
from gi.repository import GdkPixbuf
from gi.repository import GLib
from gi.repository import Gio
from blueman.gui.GtkAnimation import AnimBase
import logging

OPACITY_START = 0.7


class Fade(AnimBase):
    def __init__(self, window: Gtk.Window) -> None:
        super().__init__(state=OPACITY_START)
        self.window = window

    def state_changed(self, state: float) -> None:
        self.window.props.opacity = state


class _NotificationDialog(Gtk.MessageDialog):
    def __init__(self, summary: str, message: str, _timeout: int = -1,
                 actions: Optional[Iterable[Tuple[str, str]]] = None,
                 actions_cb: Optional[Callable[[str], None]] = None, icon_name: Optional[str] = None,
                 image_data: Optional[GdkPixbuf.Pixbuf] = None) -> None:
        super().__init__(parent=None, type=Gtk.MessageType.QUESTION,
                         buttons=Gtk.ButtonsType.NONE, text=None)

        self.set_name("NotificationDialog")
        i = 100
        self.actions_supported = True
        self.actions: Dict[int, str] = {}
        self.callback = actions_cb
        if actions:
            for a in actions:
                action_id = a[0]
                action_name = a[1]

                self.actions[i] = action_id
                self.add_button(action_name, i)
                i += 1

        self.actions[Gtk.ResponseType.DELETE_EVENT] = "close"

        self.props.secondary_use_markup = True
        self.resize(350, 50)

        self.fader = Fade(self)

        self.props.skip_taskbar_hint = False

        self.props.title = summary
        self.props.text = summary
        self.props.secondary_text = message

        self.props.window_position = Gtk.WindowPosition.CENTER

        if icon_name:
            self.set_icon_from_icon_name(icon_name, 48)
        elif image_data:
            self.set_icon_from_pixbuf(image_data)

        self.connect("response", self.dialog_response)
        self.props.icon_name = "blueman"

        self.entered = False

        def on_enter(_widget: "_NotificationDialog", _event: Gdk.Event) -> bool:
            if self.get_window() == Gdk.Window.at_pointer()[0] or not self.entered:
                self.fader.animate(start=self.fader.get_state(), end=1.0, duration=500)
                self.entered = True
            return False

        def on_leave(_widget: "_NotificationDialog", _event: Gdk.Event) -> bool:
            if not Gdk.Window.at_pointer():
                self.entered = False
                self.fader.animate(start=self.fader.get_state(), end=OPACITY_START, duration=500)
            return False

        self.connect("enter-notify-event", on_enter)
        self.connect("leave-notify-event", on_leave)

    def dialog_response(self, _dialog: Gtk.Dialog, response: int) -> None:
        if self.callback:
            self.callback(self.actions[response])
        self.hide()

    def show(self) -> None:
        self.set_opacity(OPACITY_START)
        self.present()
        self.set_opacity(OPACITY_START)

    def close(self) -> None:
        self.hide()

    def add_action(self, _action_id: str, _label: str, _callback: Optional[Callable[[str], None]] = None) -> None:
        logging.warning("stub")

    def set_icon_from_pixbuf(self, pixbuf: GdkPixbuf.Pixbuf) -> None:
        im = Gtk.Image.new_from_pixbuf(pixbuf)
        self.set_image(im)
        im.show()

    def set_icon_from_icon_name(self, icon_name: str, size: int) -> None:
        im = Gtk.Image(icon_name=icon_name, pixel_size=size)
        self.set_image(im)
        im.show()


class _NotificationBubble(Gio.DBusProxy):
    def __init__(self, summary: str, message: str, timeout: int = -1,
                 actions: Optional[Iterable[Tuple[str, str]]] = None,
                 actions_cb: Optional[Callable[[str], None]] = None, icon_name: Optional[str] = None,
                 image_data: Optional[GdkPixbuf.Pixbuf] = None) -> None:
        super().__init__(
            g_name='org.freedesktop.Notifications',
            g_interface_name='org.freedesktop.Notifications',
            g_object_path='/org/freedesktop/Notifications',
            g_bus_type=Gio.BusType.SESSION,
            g_flags=Gio.DBusProxyFlags.NONE)

        self.init()

        self._app_name = 'blueman'
        self._app_icon = ''
        self._actions: List[str] = []
        self._callbacks: Dict[str, Callable[[str], None]] = {}
        self._hints: Dict[str, GLib.Variant] = {}

        # hint : (variant format, spec version)
        self._supported_hints = {
            'action-icons': ('b', 1.2),
            'category': ('s', 0),
            'desktop-entry': ('s', 0),
            'image-data': ('(iiibiiay)', 1.2),
            'image_data': ('(iiibiiay)', 1.1),
            'image-path': ('s', 1.2),
            'image_path': ('s', 1.1),
            'icon_data': ('(iiibiiay)', 0),
            'resident': ('b', 1.2),
            'sound-file': ('s', 0),
            'sound-name': ('s', 0),
            'suppress-sound': ('b', 0),
            'transient': ('b', 1.2),
            'x': ('i', 0),
            'y': ('i', 0),
            'urgency': ('y', 0)}

        self._body = message
        self._summary = summary
        self._timeout = timeout
        self._return_id = None

        if icon_name:
            self._app_icon = icon_name
        elif image_data:
            w = image_data.props.width
            h = image_data.props.height
            stride = image_data.props.rowstride
            alpha = image_data.props.has_alpha
            bits = image_data.props.bits_per_sample
            channel = image_data.props.n_channels
            data = image_data.get_pixels()
            supported_spec = float(self.server_information[-1])
            if supported_spec < 1.1:
                key = 'icon_data'
            elif supported_spec < 1.2:
                key = 'image_data'
            elif supported_spec >= 1.2:
                key = 'image-data'
            else:
                raise ValueError('Not supported by server')

            self.set_hint(key, (w, h, stride, alpha, bits, channel, data))

        if actions:
            for action in actions:
                self.add_action(action[0], action[1], actions_cb)

        self._capabilities = self.GetCapabilities()

    @property
    def server_information(self) -> Tuple[str, str, str, str]:
        info: Tuple[str, str, str, str] = self.GetServerInformation()
        return info

    @property
    def actions_supported(self) -> bool:
        return 'actions' in self._capabilities

    def set_hint(self, key: str, val: object) -> None:
        if key not in self._supported_hints:
            raise ValueError('Unsupported hint')
        fmt, spec_version = self._supported_hints[key]
        if spec_version > float(self.server_information[-1]):
            raise ValueError('Not supported by server')

        param = GLib.Variant(fmt, val)
        self._hints[key] = param

    def remove_hint(self, key: str) -> None:
        del (self._hints[key])

    def clear_hints(self) -> None:
        self._hints = {}

    def add_action(self, action_id: str, label: str, callback: Optional[Callable[[str], None]] = None) -> None:
        self._actions.extend([action_id, label])
        if callback:
            self._callbacks[action_id] = callback

    def clear_actions(self) -> None:
        self._actions = []
        self._callbacks = {}

    def do_g_signal(self, _sender_name: str, signal_name: str, params: GLib.Variant) -> None:
        notif_id, signal_val = params.unpack()
        if notif_id != self._return_id:
            return

        logging.info(signal_val)

        if signal_name == 'NotificationClosed':
            if signal_val == 1:
                logging.debug('The notification expired.')
            elif signal_val == 2:
                logging.debug('The notification was dismissed by the user.')
            elif signal_val == 3:
                logging.debug('The notification was closed by a call to CloseNotification.')
            elif signal_val == 4:
                logging.debug('Undefined/reserved reasons.')
        elif signal_name == 'ActionInvoked':
            if signal_val in self._callbacks:
                self._callbacks[signal_val](signal_val)

    def show(self) -> None:
        replace_id = self._return_id if self._return_id else 0
        return_id = self.Notify('(susssasa{sv}i)', self._app_name, replace_id, self._app_icon,
                                self._summary, self._body, self._actions, self._hints,
                                self._timeout)
        self._return_id = return_id

    def close(self) -> None:
        param = GLib.Variant('(u)', (self._return_id,))
        self.call_sync('CloseNotification', param, Gio.DBusCallFlags.NONE, -1, None)
        self._return_id = None


def Notification(summary: str, message: str, timeout: int = -1, actions: Optional[Iterable[Tuple[str, str]]] = None,
                 actions_cb: Optional[Callable[[str], None]] = None, icon_name: Optional[str] = None,
                 image_data: Optional[GdkPixbuf.Pixbuf] = None) -> Union[_NotificationBubble, _NotificationDialog]:

    forced_fallback = not Config('org.blueman.general')['notification-daemon']
    try:
        bus = Gio.bus_get_sync(Gio.BusType.SESSION)
        caps = bus.call_sync('org.freedesktop.Notifications', '/org/freedesktop/Notifications',
                             'org.freedesktop.Notifications', 'GetCapabilities', None, None,
                             Gio.DBusCallFlags.NONE, -1, None).unpack()[0]
    except GLib.Error:
        caps = []

    if forced_fallback or 'body' not in caps or (actions and 'actions' not in caps):
        # Use fallback in the case:
        # * user does not want to use a notification daemon
        # * the notification daemon is not available
        # * we have to show actions and the notification daemon does not provide them
        klass: Type[Union[_NotificationBubble, _NotificationDialog]] = _NotificationDialog
    else:
        klass = _NotificationBubble

    return klass(summary, message, timeout, actions, actions_cb, icon_name, image_data)



from gettext import gettext as _
import logging
from typing import List, TYPE_CHECKING, Callable

import gi

from blueman.bluemantyping import GSignals

gi.require_version("Gtk", "3.0")
gi.require_version("Gdk", "3.0")
from gi.repository import Pango
from gi.repository import GObject
from gi.repository import Gtk
from gi.repository import Gdk
from gi.repository import GLib

if TYPE_CHECKING:
    from blueman.main.Manager import Blueman


class ManagerProgressbar(GObject.GObject):
    __gsignals__: GSignals = {
        'cancelled': (GObject.SignalFlags.RUN_LAST, None, ()),
    }
    __instances__: List["ManagerProgressbar"] = []

    def __init__(self, blueman: "Blueman", cancellable: bool = True, text: str = _("Connecting")) -> None:
        super().__init__()
        self.Blueman = blueman

        self.cancellable = cancellable

        self.hbox = hbox = blueman.builder.get_widget("status_data", Gtk.Box)

        self.progressbar = Gtk.ProgressBar()
        self.progressbar.set_name("ManagerProgressbar")

        self._signals: List[int] = []

        self.button = Gtk.Image(icon_name="process-stop", pixel_size=16)

        self.eventbox = eventbox = Gtk.EventBox()
        eventbox.add(self.button)
        eventbox.props.tooltip_text = _("Cancel Operation")
        eventbox.connect("enter-notify-event", self._on_enter)
        eventbox.connect("leave-notify-event", self._on_leave)
        eventbox.connect("button-press-event", self._on_clicked)

        self.progressbar.set_size_request(100, 15)
        self.progressbar.set_ellipsize(Pango.EllipsizeMode.END)
        self.progressbar.set_text(text)
        self.progressbar.set_pulse_step(0.05)

        hbox.pack_end(eventbox, True, False, 0)
        hbox.pack_end(self.progressbar, False, False, 0)

        if ManagerProgressbar.__instances__:
            logging.info(f"hiding {ManagerProgressbar.__instances__[-1]}")
            ManagerProgressbar.__instances__[-1].hide()

        self.show()
        if not self.cancellable:
            self.eventbox.props.sensitive = False

        self.pulsing = False
        self.finalized = False

        ManagerProgressbar.__instances__.append(self)

    def _get_window(self) -> Gdk.Window:
        assert self.Blueman.window is not None
        window = self.Blueman.window.get_window()
        assert window is not None
        return window

    def _on_enter(self, _evbox: Gtk.EventBox, _event: Gdk.Event) -> bool:
        c = Gdk.Cursor.new(Gdk.CursorType.HAND2)
        self._get_window().set_cursor(c)
        return False

    def _on_leave(self, _evbox: Gtk.EventBox, _event: Gdk.Event) -> bool:
        assert self.Blueman.window is not None
        self._get_window().set_cursor(None)
        return False

    def _on_clicked(self, _evbox: Gtk.EventBox, _event: Gdk.Event) -> bool:
        self.eventbox.props.sensitive = False
        self.emit("cancelled")
        return False

    def connect(self, signal: str, callback: Callable[..., None], *args: object) -> int:
        handler_id: int = super().connect(signal, callback, *args)
        self._signals.append(handler_id)
        return handler_id

    def show(self) -> None:
        if not self.Blueman.Config["show-statusbar"]:
            statusbar = self.Blueman.builder.get_widget("statusbar", Gtk.Box)
            statusbar.props.visible = True

        self.progressbar.props.visible = True
        self.eventbox.props.visible = True
        self.button.props.visible = True

    def hide(self) -> None:
        self.Blueman.Stats.hbox.show_all()
        self.progressbar.props.visible = False
        self.eventbox.props.visible = False
        self.button.props.visible = False

    def message(self, msg: str, timeout: int = 1500) -> None:
        self.stop()
        self.set_label(msg)
        self.set_cancellable(False)
        GLib.timeout_add(timeout, self.finalize)

    def finalize(self) -> bool:
        if not self.finalized:
            self.hide()
            self.stop()
            self._get_window().set_cursor(None)
            self.hbox.remove(self.eventbox)
            self.hbox.remove(self.progressbar)
            # self.hbox.remove(self.seperator)
            self.finalized = True

            if ManagerProgressbar.__instances__[-1] == self:
                ManagerProgressbar.__instances__.pop()
                # remove all finalized instances
                for inst in reversed(ManagerProgressbar.__instances__):
                    if inst.finalized:
                        ManagerProgressbar.__instances__.pop()
                    else:
                        # show last active progress bar
                        inst.show()
                        break

            if not ManagerProgressbar.__instances__:
                if not self.Blueman.Config["show-statusbar"]:
                    statusbar = self.Blueman.builder.get_widget("statusbar", Gtk.Box)
                    statusbar.props.visible = False

            for sig in self._signals:
                if self.handler_is_connected(sig):
                    self.disconnect(sig)
            self._signals = []
        return False

    def set_cancellable(self, b: bool, hide: bool = False) -> None:
        if b:
            self.eventbox.props.visible = True
            self.eventbox.props.sensitive = True
        else:
            if hide:
                self.eventbox.props.visible = False
            else:
                self.eventbox.props.sensitive = False

    def set_label(self, label: str) -> None:
        self.progressbar.props.text = label

    def fraction(self, frac: float) -> None:
        if not self.finalized:
            self.progressbar.set_fraction(frac)

    def started(self) -> bool:
        return self.pulsing

    def start(self) -> None:
        def pulse() -> bool:
            self.progressbar.pulse()
            return self.pulsing

        if not self.pulsing:
            self.pulsing = True
            GLib.timeout_add(41, pulse)

    def stop(self) -> None:
        self.pulsing = False
        self.progressbar.set_fraction(0.0)



from blueman.bluez.obex.Base import Base


class Session(Base):
    _interface_name = 'org.bluez.obex.Session1'

    def __init__(self, obj_path: str):
        super().__init__(obj_path=obj_path)

    @property
    def address(self) -> str:
        dest: str = self.get('Destination')
        return dest

    @property
    def root(self) -> str:
        root: str = self.get('Root')
        return root



from gettext import gettext as _

from blueman.plugins.AppletPlugin import AppletPlugin
from blueman.main.applet.BluezAgent import BluezAgent


class AuthAgent(AppletPlugin):
    __description__ = _("Provides passkey, authentication services for BlueZ daemon")
    __icon__ = "dialog-password"
    __author__ = "Walmis"
    __depends__ = ["StatusIcon"]

    _agent = None

    def on_unload(self) -> None:
        if self._agent:
            self._agent.unregister_agent()
            self._agent = None

    def on_manager_state_changed(self, state: bool) -> None:
        if state:
            self._agent = BluezAgent()
            self._agent.register_agent()
        else:
            # At this point bluez already called Release on the agent
            self._agent = None



import logging
import weakref
from typing import Dict, Callable, List, Tuple

from gi.repository import GObject, Gio

from blueman.bluez.obex.Transfer import Transfer
from blueman.gobject import SingletonGObjectMeta
from blueman.bluemantyping import GSignals


class Manager(GObject.GObject, metaclass=SingletonGObjectMeta):
    __gsignals__: GSignals = {
        'session-added': (GObject.SignalFlags.NO_HOOKS, None, (str,)),
        'session-removed': (GObject.SignalFlags.NO_HOOKS, None, (str,)),
        'transfer-started': (GObject.SignalFlags.NO_HOOKS, None, (str,)),
        'transfer-completed': (GObject.SignalFlags.NO_HOOKS, None, (str, bool)),
    }

    connect_signal = GObject.GObject.connect
    disconnect_signal = GObject.GObject.disconnect

    __bus_name = 'org.bluez.obex'

    def __init__(self) -> None:
        super().__init__()
        self.__transfers: Dict[str, Tuple[Transfer, Tuple[int, ...]]] = {}

        self._object_manager = Gio.DBusObjectManagerClient.new_for_bus_sync(
            Gio.BusType.SESSION, Gio.DBusObjectManagerClientFlags.NONE,
            self.__bus_name, '/', None, None, None)

        self._manager_handlerids: List[int] = []
        self._manager_handlerids.append(self._object_manager.connect('object-added', self._on_object_added))
        self._manager_handlerids.append(self._object_manager.connect('object-removed', self._on_object_removed))

        weakref.finalize(self, self._on_delete)

    def _on_delete(self) -> None:
        for handlerid in self._manager_handlerids:
            self._object_manager.disconnect(handlerid)
        self._manager_handlerids = []

    def _on_object_added(self, _object_manager: Gio.DBusObjectManager, dbus_object: Gio.DBusObject) -> None:
        session_proxy = dbus_object.get_interface('org.bluez.obex.Session1')
        transfer_proxy = dbus_object.get_interface('org.bluez.obex.Transfer1')
        object_path = dbus_object.get_object_path()

        if transfer_proxy:
            logging.info(object_path)
            transfer = Transfer(obj_path=object_path)
            chandlerid = transfer.connect_signal('completed', self._on_transfer_completed, True)
            ehandlerid = transfer.connect_signal('error', self._on_transfer_completed, False)
            self.__transfers[object_path] = (transfer, (chandlerid, ehandlerid))
            self.emit('transfer-started', object_path)

        if session_proxy:
            logging.info(object_path)
            self.emit('session-added', object_path)

    def _on_object_removed(self, _object_manager: Gio.DBusObjectManager, dbus_object: Gio.DBusObject) -> None:
        session_proxy = dbus_object.get_interface('org.bluez.obex.Session1')
        transfer_proxy = dbus_object.get_interface('org.bluez.obex.Transfer1')
        object_path = dbus_object.get_object_path()

        if transfer_proxy and object_path in self.__transfers:
            logging.info(object_path)
            transfer, handlerids = self.__transfers.pop(object_path)

            for handlerid in handlerids:
                transfer.disconnect_signal(handlerid)

        if session_proxy:
            logging.info(object_path)
            self.emit('session-removed', object_path)

    def _on_transfer_completed(self, transfer: Transfer, success: bool) -> None:
        transfer_path = transfer.get_object_path()

        logging.info(f"{transfer_path} {success}")
        self.emit('transfer-completed', transfer_path, success)

    @classmethod
    def watch_name_owner(
        cls,
        appeared_handler: Callable[[Gio.DBusConnection, str, str], None],
        vanished_handler: Callable[[Gio.DBusConnection, str], None],
    ) -> None:
        Gio.bus_watch_name(Gio.BusType.SESSION, cls.__bus_name, Gio.BusNameWatcherFlags.AUTO_START,
                           appeared_handler, vanished_handler)



import datetime
from gettext import gettext as _
from tempfile import NamedTemporaryFile
from typing import List

from blueman.Functions import create_menuitem, launch
from blueman.bluez.Device import Device
from blueman.gui.manager.ManagerDeviceMenu import MenuItemsProvider, ManagerDeviceMenu, DeviceMenuItem
from blueman.main.Builder import Builder
from blueman.plugins.ManagerPlugin import ManagerPlugin

import gi
gi.require_version("Gtk", "3.0")
from gi.repository import Gtk


def send_note_cb(dialog: Gtk.Dialog, response_id: int, device_address: str, text_view: Gtk.Entry) -> None:
    text = text_view.get_buffer().props.text
    dialog.destroy()
    if response_id == Gtk.ResponseType.CANCEL:
        return

    date = datetime.datetime.now().strftime('%Y%m%dT%H%M00')
    data = ('BEGIN:VNOTE \n'
            'VERSION:1.1 \n'
            'BODY;CHARSET=UTF-8: %s \n'
            'DCREATED:%s \n'
            'LAST-MODIFIED:%s \n'
            'CLASS:PUBLIC \n'
            'X-IRMC-LUID:000001000000 \n'
            'END:VNOTE \n' % (' '.join(text.splitlines()), date, date))

    tempfile = NamedTemporaryFile(suffix='.vnt', prefix='note', delete=False)
    tempfile.write(data.encode('utf-8'))
    tempfile.close()
    launch(f"blueman-sendto --delete --device={device_address}", paths=[tempfile.name])


def send_note(device: Device, parent: Gtk.Window) -> None:
    builder = Builder("note.ui")
    dialog = builder.get_widget("dialog", Gtk.Dialog)
    dialog.set_transient_for(parent)
    dialog.props.icon_name = 'blueman'
    note = builder.get_widget("note", Gtk.Entry)
    dialog.connect('response', send_note_cb, device['Address'], note)
    dialog.present()


class Notes(ManagerPlugin, MenuItemsProvider):
    def on_request_menu_items(self, manager_menu: ManagerDeviceMenu, device: Device) -> List[DeviceMenuItem]:
        item = create_menuitem(_("Send _note"), "dialog-information")
        item.props.tooltip_text = _("Send a text note")
        _window = manager_menu.get_toplevel()
        assert isinstance(_window, Gtk.Window)
        window = _window  # https://github.com/python/mypy/issues/2608
        item.connect('activate', lambda x: send_note(device, window))
        return [DeviceMenuItem(item, DeviceMenuItem.Group.ACTIONS, 500)]



from gettext import gettext as _
import logging
from typing import List, Any

from gi.repository import GLib

from blueman.bluez.Network import AnyNetwork, Network
from blueman.gui.Notification import Notification
from blueman.plugins.AppletPlugin import AppletPlugin
from blueman.main.DBusProxies import Mechanism


class DhcpClient(AppletPlugin):
    __description__ = _("Provides a basic dhcp client for Bluetooth PAN connections.")
    __icon__ = "network-workgroup"
    __author__ = "Walmis"

    _any_network = None

    def on_load(self) -> None:
        self._any_network = AnyNetwork()
        self._any_network.connect_signal('property-changed', self._on_network_prop_changed)

        self.quering: List[str] = []

        self._add_dbus_method("DhcpClient", ("s",), "", self.dhcp_acquire)

    def on_unload(self) -> None:
        del self._any_network

    def _on_network_prop_changed(self, _network: AnyNetwork, key: str, value: Any, object_path: str) -> None:
        if key == "Interface":
            if value != "":
                self.dhcp_acquire(object_path)

    def dhcp_acquire(self, object_path: str) -> None:
        device = Network(obj_path=object_path)["Interface"]

        if device not in self.quering:
            self.quering.append(device)
        else:
            return

        if device != "":
            def reply(_obj: Mechanism, result: str, _user_data: None) -> None:
                logging.info(result)
                Notification(_("Bluetooth Network"),
                             _("Interface %(0)s bound to IP address %(1)s") % {"0": device, "1": result},
                             icon_name="network-workgroup").show()

                self.quering.remove(device)

            def err(_obj: Mechanism, result: GLib.Error, _user_data: None) -> None:
                logging.warning(result)
                Notification(_("Bluetooth Network"), _("Failed to obtain an IP address on %s") % device,
                             icon_name="network-workgroup").show()

                self.quering.remove(device)

            Notification(_("Bluetooth Network"), _("Trying to obtain an IP address on %s\nPlease wait…" % device),
                         icon_name="network-workgroup").show()

            m = Mechanism()
            m.DhcpClient('(s)', object_path, result_handler=reply, error_handler=err, timeout=120 * 1000)



from typing import Optional, Callable

from gi.repository import GLib

from blueman.bluez.AnyBase import AnyBase
from blueman.bluez.Base import Base
from blueman.bluez.Device import Device
from blueman.bluez.errors import BluezDBusException


class Adapter(Base):
    _interface_name = 'org.bluez.Adapter1'

    def __init__(self, obj_path: str):
        super().__init__(obj_path=obj_path)

    def start_discovery(self, error_handler: Optional[Callable[[BluezDBusException], None]] = None) -> None:
        self._call('StartDiscovery', error_handler=error_handler)

    def stop_discovery(self) -> None:
        self._call('StopDiscovery')

    def remove_device(self, device: Device) -> None:
        param = GLib.Variant('(o)', (device.get_object_path(),))
        self._call('RemoveDevice', param)

    def get_name(self) -> str:
        name: str = self['Alias']
        return name

    def set_name(self, name: str) -> None:
        self.set('Alias', name)


class AnyAdapter(AnyBase):
    def __init__(self) -> None:
        super().__init__('org.bluez.Adapter1')



from gettext import gettext as _
import logging
from typing import Dict, List, TYPE_CHECKING, Mapping, Sequence

from blueman.bluez.Device import Device
from blueman.plugins.ManagerPlugin import ManagerPlugin
from blueman.main.PulseAudioUtils import PulseAudioUtils, EventType
from blueman.gui.manager.ManagerDeviceMenu import ManagerDeviceMenu, MenuItemsProvider, DeviceMenuItem
from blueman.gui.MessageArea import MessageArea
from blueman.Functions import create_menuitem
from blueman.Sdp import AUDIO_SOURCE_SVCLASS_ID, AUDIO_SINK_SVCLASS_ID, ServiceUUID

import gi
gi.require_version("Gtk", "3.0")
from gi.repository import Gtk


if TYPE_CHECKING:
    from blueman.main.PulseAudioUtils import CardInfo  # noqa: F401


class PulseAudioProfile(ManagerPlugin, MenuItemsProvider):
    def on_load(self) -> None:
        self.devices: Dict[str, "CardInfo"] = {}

        self.deferred: List[Device] = []

        pa = PulseAudioUtils()
        pa.connect("event", self.on_pa_event)
        pa.connect("connected", self.on_pa_ready)

    def on_pa_ready(self, _utils: PulseAudioUtils) -> None:
        logging.info("connected")
        for dev in self.deferred:
            self.regenerate_with_device(dev['Address'])

        self.deferred = []

    # updates all menu instances with the following device address
    def regenerate_with_device(self, device_addr: str) -> None:
        for inst in ManagerDeviceMenu.__instances__:
            if inst.SelectedDevice['Address'] == device_addr and not inst.is_popup:
                inst.generate()

    def on_pa_event(self, utils: PulseAudioUtils, event: int, idx: int) -> None:
        logging.debug(f"{event} {idx}")

        def get_card_cb(card: "CardInfo") -> None:
            drivers = ("module-bluetooth-device.c",
                       "module-bluez4-device.c",
                       "module-bluez5-device.c")

            if card["driver"] in drivers:
                self.devices[card["proplist"]["device.string"]] = card
                self.regenerate_with_device(card["proplist"]["device.string"])

        if event & EventType.CARD:
            logging.info("card")
            if event & EventType.CHANGE:
                logging.info("change")
                utils.get_card(idx, get_card_cb)
            elif event & EventType.REMOVE:
                logging.info("remove")
            else:
                logging.info("add")
                utils.get_card(idx, get_card_cb)

    def query_pa(self, device: Device, item: Gtk.MenuItem) -> None:
        def list_cb(cards: Mapping[str, "CardInfo"]) -> None:
            for c in cards.values():
                if c["proplist"]["device.string"] == device['Address']:
                    self.devices[device['Address']] = c
                    self.generate_menu(device, item)
                    return

        pa = PulseAudioUtils()
        pa.list_cards(list_cb)

    def on_selection_changed(self, item: Gtk.CheckMenuItem, device: Device, profile: str) -> None:
        if item.get_active():
            pa = PulseAudioUtils()

            c = self.devices[device['Address']]

            def on_result(res: int) -> None:
                if not res:
                    MessageArea.show_message(_("Failed to change profile to %s" % profile))

            pa.set_card_profile(c["index"], profile, on_result)

    def generate_menu(self, device: Device, item: Gtk.MenuItem) -> None:
        info = self.devices[device['Address']]
        group: Sequence[Gtk.RadioMenuItem] = []

        sub = Gtk.Menu()

        if info:
            for profile in info["profiles"]:
                i = Gtk.RadioMenuItem.new_with_label(group, profile["description"])
                group = i.get_group()

                if profile["name"] == info["active_profile"]:
                    i.set_active(True)

                i.connect("toggled", self.on_selection_changed,
                          device, profile["name"])

                sub.append(i)
                i.show()

            item.set_submenu(sub)
            item.show()

    def on_request_menu_items(self, manager_menu: ManagerDeviceMenu, device: Device) -> List[DeviceMenuItem]:
        audio_source = False
        for uuid in device['UUIDs']:
            if ServiceUUID(uuid).short_uuid in (AUDIO_SOURCE_SVCLASS_ID, AUDIO_SINK_SVCLASS_ID):
                audio_source = True
                break

        if device['Connected'] and audio_source:

            pa = PulseAudioUtils()
            if not pa.connected:
                self.deferred.append(device)
                return []

            item = create_menuitem(_("Audio Profile"), "audio-card")
            item.props.tooltip_text = _("Select audio profile for PulseAudio")

            if not device['Address'] in self.devices:
                self.query_pa(device, item)
            else:
                self.generate_menu(device, item)

        else:
            return []

        return [DeviceMenuItem(item, DeviceMenuItem.Group.ACTIONS, 300)]



from typing import Dict, Optional, TYPE_CHECKING, Iterable, Mapping, Callable, Tuple, Union, Collection, List, Any

import gi
gi.require_version("Gtk", "3.0")
from gi.repository import Gtk


if TYPE_CHECKING:
    from typing_extensions import TypedDict

    class _ListDataDictBase(TypedDict):
        id: str
        type: type

    class ListDataDict(_ListDataDictBase, total=False):
        renderer: Gtk.CellRenderer
        render_attrs: Mapping[str, int]
        view_props: Mapping[str, object]
        celldata_func: Tuple[Callable[[Gtk.TreeViewColumn, Gtk.CellRenderer, Gtk.TreeModel, Gtk.TreeIter, Any],
                                      None], Any]
else:
    ListDataDict = dict


# noinspection PyAttributeOutsideInit
class GenericList(Gtk.TreeView):
    def __init__(self, data: Iterable[ListDataDict], headers_visible: bool = True, visible: bool = False) -> None:
        super().__init__(headers_visible=headers_visible, visible=visible)
        self.set_name("GenericList")
        self.selection = self.get_selection()
        self._load(data)

    def _load(self, data: Iterable[ListDataDict]) -> None:
        self.ids: Dict[str, int] = {}
        self.columns: Dict[str, Gtk.TreeViewColumn] = {}

        types = [row["type"] for row in data]

        self.liststore = Gtk.ListStore(*types)
        self.set_model(self.liststore)

        for i, row in enumerate(data):
            self.ids[row["id"]] = i

            if "renderer" not in row:
                continue

            column = Gtk.TreeViewColumn()
            column.pack_start(row["renderer"], True)
            column.set_attributes(row["renderer"], **row["render_attrs"])

            if "view_props" in row:
                column.set_properties(**row["view_props"])

            if "celldata_func" in row:
                func, user_data = row["celldata_func"]
                column.set_cell_data_func(row["renderer"], func, user_data)

            self.columns[row["id"]] = column
            self.append_column(column)

    def selected(self) -> Optional[Gtk.TreeIter]:
        (model, tree_iter) = self.selection.get_selected()
        return tree_iter

    def delete(self, iterid: Union[Gtk.TreeIter, Gtk.TreePath, int, str]) -> bool:
        if isinstance(iterid, Gtk.TreeIter):
            tree_iter: Optional[Gtk.TreeIter] = iterid
        else:
            tree_iter = self.get_iter(iterid)

        if tree_iter is None:
            return False
        if self.liststore.iter_is_valid(tree_iter):
            self.liststore.remove(tree_iter)
            return True
        else:
            return False

    def _add(self, **columns: object) -> Collection[object]:
        items: Dict[int, object] = {}
        for k, v in self.ids.items():
            items[v] = None

        for k, val in columns.items():
            if k in self.ids:
                items[self.ids[k]] = val
            else:
                raise Exception(f"Invalid key {k}")

        return items.values()

    def append(self, **columns: object) -> Gtk.TreeIter:
        vals = self._add(**columns)
        return self.liststore.append(vals)

    def prepend(self, **columns: object) -> Gtk.TreeIter:
        vals = self._add(**columns)
        return self.liststore.prepend(vals)

    def get_conditional(self, **cols: object) -> List[int]:
        ret = []
        matches = 0
        for i in range(len(self.liststore)):
            row = self.get(i)
            for k, v in cols.items():
                if row[k] == v:
                    matches += 1

            if matches == len(cols):
                ret.append(i)

            matches = 0

        return ret

    def set(self, iterid: Union[Gtk.TreeIter, int, str], **cols: object) -> None:
        if isinstance(iterid, Gtk.TreeIter):
            tree_iter: Optional[Gtk.TreeIter] = iterid
        else:
            tree_iter = self.get_iter(iterid)

        if tree_iter is not None:
            for k, v in cols.items():
                self.liststore.set(tree_iter, self.ids[k], v)

    def get(self, iterid: Union[Gtk.TreeIter, Gtk.TreePath, int, str], *items: str) -> Dict[str, Any]:
        ret = {}

        if iterid is not None:
            if isinstance(iterid, Gtk.TreeIter):
                tree_iter: Optional[Gtk.TreeIter] = iterid
            else:
                tree_iter = self.get_iter(iterid)
            assert tree_iter is not None
            if len(items) == 0:
                for k, v in self.ids.items():
                    ret[k] = self.liststore.get(tree_iter, v)[0]
            else:
                for i in range(len(items)):
                    if items[i] in self.ids:
                        ret[items[i]] = self.liststore.get(tree_iter, self.ids[items[i]])[0]
        else:
            return {}

        return ret

    def get_iter(self, path: Optional[Union[Gtk.TreePath, int, str]]) -> Optional[Gtk.TreeIter]:
        if path is None:
            return None

        try:
            return self.liststore.get_iter(path)
        except ValueError:
            return None

    def clear(self) -> None:
        self.liststore.clear()

    def compare(self, iter_a: Optional[Gtk.TreeIter], iter_b: Optional[Gtk.TreeIter]) -> bool:
        if iter_a is not None and iter_b is not None:
            model = self.get_model()
            assert model is not None
            return model.get_path(iter_a) == model.get_path(iter_b)
        else:
            return False



from gettext import gettext as _

from blueman.plugins.AppletPlugin import AppletPlugin

# Check if Appindicator is available and raise ImportError
from gi import require_version

from blueman.plugins.applet.StatusIcon import StatusIconImplementationProvider

try:
    require_version('AppIndicator3', '0.1')
except ValueError:
    raise ImportError("AppIndicator3 not found")


class AppIndicator(AppletPlugin, StatusIconImplementationProvider):
    __description__ = _("Uses libappindicator to show a statusicon")
    __icon__ = "blueman-tray"
    __author__ = "Walmis"
    __depends__ = ['StatusIcon']

    def on_query_status_icon_implementation(self) -> str:
        return 'AppIndicator'



from typing import TYPE_CHECKING

from blueman.plugins.BasePlugin import BasePlugin

if TYPE_CHECKING:
    from blueman.main.Manager import Blueman


class ManagerPlugin(BasePlugin):
    def __init__(self, parent: "Blueman"):
        super().__init__()
        self.parent = parent

    def on_unload(self) -> None:
        pass



from gettext import gettext as _
from typing import Optional, TYPE_CHECKING, List, Any, cast, Callable, Set
import html
import logging
import cairo
import os

from blueman.bluez.Battery import Battery
from blueman.bluez.Device import Device
from blueman.gui.DeviceList import DeviceList
from blueman.DeviceClass import get_minor_class, get_major_class, gatt_appearance_to_name
from blueman.gui.GenericList import ListDataDict
from blueman.gui.manager.ManagerDeviceMenu import ManagerDeviceMenu
from blueman.Constants import PIXMAP_PATH
from blueman.Functions import launch
from blueman.Sdp import ServiceUUID, OBEX_OBJPUSH_SVCLASS_ID, BATTERY_SERVICE_SVCLASS_ID
from blueman.gui.GtkAnimation import TreeRowFade, CellFade, AnimBase
from blueman.main.Config import Config
from _blueman import ConnInfoReadError, conn_info

import gi
gi.require_version("Gtk", "3.0")
from gi.repository import Gtk
from gi.repository import GLib
from gi.repository import Gdk
from gi.repository import GdkPixbuf
from gi.repository import Pango

if TYPE_CHECKING:
    from blueman.main.Manager import Blueman


class ManagerDeviceList(DeviceList):
    def __init__(self, adapter: Optional[str] = None, inst: Optional["Blueman"] = None) -> None:
        cr = Gtk.CellRendererText()
        cr.props.ellipsize = Pango.EllipsizeMode.END
        tabledata: List[ListDataDict] = [
            # device picture
            {"id": "device_surface", "type": str, "renderer": Gtk.CellRendererPixbuf(),
             "render_attrs": {}, "celldata_func": (self._set_cell_data, None)},
            # device caption
            {"id": "caption", "type": str, "renderer": cr,
             "render_attrs": {"markup": 1}, "view_props": {"expand": True}},
            {"id": "battery_pb", "type": GdkPixbuf.Pixbuf, "renderer": Gtk.CellRendererPixbuf(),
             "render_attrs": {}, "view_props": {"spacing": 0},
             "celldata_func": (self._set_cell_data, "battery")},
            {"id": "rssi_pb", "type": GdkPixbuf.Pixbuf, "renderer": Gtk.CellRendererPixbuf(),
             "render_attrs": {}, "view_props": {"spacing": 0},
             "celldata_func": (self._set_cell_data, "rssi")},
            {"id": "lq_pb", "type": GdkPixbuf.Pixbuf, "renderer": Gtk.CellRendererPixbuf(),
             "render_attrs": {}, "view_props": {"spacing": 0},
             "celldata_func": (self._set_cell_data, "lq")},
            {"id": "tpl_pb", "type": GdkPixbuf.Pixbuf, "renderer": Gtk.CellRendererPixbuf(),
             "render_attrs": {}, "view_props": {"spacing": 0},
             "celldata_func": (self._set_cell_data, "tpl")},
            {"id": "alias", "type": str},  # used for quick access instead of device.GetProperties
            {"id": "connected", "type": bool},  # used for quick access instead of device.GetProperties
            {"id": "paired", "type": bool},  # used for quick access instead of device.GetProperties
            {"id": "trusted", "type": bool},  # used for quick access instead of device.GetProperties
            {"id": "objpush", "type": bool},  # used to set Send File button
            {"id": "battery", "type": float},
            {"id": "rssi", "type": float},
            {"id": "lq", "type": float},
            {"id": "tpl", "type": float},
            {"id": "icon_info", "type": Gtk.IconInfo},
            {"id": "cell_fader", "type": CellFade},
            {"id": "row_fader", "type": TreeRowFade},
            {"id": "initial_anim", "type": bool},
        ]
        super().__init__(adapter, tabledata)
        self.set_name("ManagerDeviceList")
        self.set_headers_visible(False)
        self.props.has_tooltip = True
        self.Blueman = inst

        self._monitored_devices: Set[str] = set()

        self.Config = Config("org.blueman.general")
        self.Config.connect('changed', self._on_settings_changed)
        # Set the correct sorting
        self._on_settings_changed(self.Config, "sort-by")
        self._on_settings_changed(self.Config, "sort-type")

        self.connect("query-tooltip", self.tooltip_query)
        self.tooltip_row: Optional[Gtk.TreePath] = None
        self.tooltip_col: Optional[Gtk.TreeViewColumn] = None

        self.connect("popup-menu", self._on_popup_menu)
        self.connect("button_press_event", self.on_event_clicked)
        self.connect("button_release_event", self.on_event_clicked)

        self.menu: Optional[ManagerDeviceMenu] = None

        self.connect("drag_data_received", self.drag_recv)
        self.connect("drag-motion", self.drag_motion)

        Gtk.Widget.drag_dest_set(self, Gtk.DestDefaults.ALL, [], Gdk.DragAction.COPY | Gdk.DragAction.DEFAULT)
        Gtk.Widget.drag_dest_add_uri_targets(self)

        self.set_search_equal_func(self.search_func)

    def _on_settings_changed(self, settings: Config, key: str) -> None:
        if key in ('sort-by', 'sort-order'):
            sort_by = settings['sort-by']
            sort_order = settings['sort-order']

            if sort_order == 'ascending':
                sort_type = Gtk.SortType.ASCENDING
            else:
                sort_type = Gtk.SortType.DESCENDING

            column_id = self.ids.get(sort_by)

            if column_id:
                self.liststore.set_sort_column_id(column_id, sort_type)

    def on_icon_theme_changed(self, _icon_them: Gtk.IconTheme) -> None:
        for row in self.liststore:
            device = self.get(row.iter, "device")["device"]
            self.row_setup_event(row.iter, device)

    def search_func(self, model: Gtk.TreeModel, column: int, key: str, tree_iter: Gtk.TreeIter) -> bool:
        row = self.get(tree_iter, "caption")
        if key.lower() in row["caption"].lower():
            return False
        logging.info(f"{model} {column} {key} {tree_iter}")
        return True

    def drag_recv(self, _widget: Gtk.Widget, context: Gdk.DragContext, x: int, y: int, selection: Gtk.SelectionData,
                  _info: int, time: int) -> None:

        uris = list(selection.get_uris())

        context.finish(True, False, time)

        path = self.get_path_at_pos(x, y)
        if path:
            tree_iter = self.get_iter(path[0])
            assert tree_iter is not None
            device = self.get(tree_iter, "device")["device"]
            command = f"blueman-sendto --device={device['Address']}"

            launch(command, paths=uris, name=_("File Sender"))
            context.finish(True, False, time)
        else:
            context.finish(False, False, time)

    def drag_motion(self, _widget: Gtk.Widget, drag_context: Gdk.DragContext, x: int, y: int, timestamp: int) -> bool:
        result = self.get_path_at_pos(x, y)
        if result is not None:
            path = result[0]
            assert path is not None
            if not self.selection.path_is_selected(path):
                tree_iter = self.get_iter(path)
                assert tree_iter is not None
                has_obj_push = self._has_objpush(self.get(tree_iter, "device")["device"])
                if has_obj_push:
                    Gdk.drag_status(drag_context, Gdk.DragAction.COPY, timestamp)
                    self.set_cursor(path)
                    return True
                else:
                    Gdk.drag_status(drag_context, Gdk.DragAction.DEFAULT, timestamp)
                    return False
            return False
        else:
            Gdk.drag_status(drag_context, Gdk.DragAction.DEFAULT, timestamp)
            return False

    def _on_popup_menu(self, _widget: Gtk.Widget) -> bool:
        if self.Blueman is None:
            return False

        if self.menu is None:
            self.menu = ManagerDeviceMenu(self.Blueman)

        window = self.get_window()
        assert window is not None
        selected = self.selected()
        assert selected is not None
        rect = self.get_cell_area(self.liststore.get_path(selected), self.get_column(1))
        self.menu.popup_at_rect(window, rect, Gdk.Gravity.CENTER, Gdk.Gravity.NORTH)

        return True

    def on_event_clicked(self, _widget: Gtk.Widget, event: Gdk.Event) -> bool:
        if event.type not in (Gdk.EventType._2BUTTON_PRESS, Gdk.EventType.BUTTON_PRESS):
            return False

        path = self.get_path_at_pos(int(cast(Gdk.EventButton, event).x), int(cast(Gdk.EventButton, event).y))
        if path is None:
            return False

        assert path[0] is not None
        row = self.get(path[0], "device", "connected")
        if not row:
            return False

        if self.Blueman is None:
            return False

        if self.menu is None:
            self.menu = ManagerDeviceMenu(self.Blueman)

        if event.type == Gdk.EventType._2BUTTON_PRESS and cast(Gdk.EventButton, event).button == 1:
            if self.menu.show_generic_connect_calc(row["device"]['UUIDs']):
                if row["connected"]:
                    self.menu.disconnect_service(row["device"])
                else:
                    self.menu.connect_service(row["device"])

        if event.type == Gdk.EventType.BUTTON_PRESS and cast(Gdk.EventButton, event).button == 3:
            self.menu.popup_at_pointer(event)

        return False

    def get_icon_info(self, icon_name: str, size: int = 48, fallback: bool = True) -> Optional[Gtk.IconInfo]:
        if icon_name is None and not fallback:
            return None
        elif icon_name is None and fallback:
            icon_name = "image-missing"

        icon_info = self.icon_theme.lookup_icon_for_scale(icon_name, size, self.get_scale_factor(),
                                                          Gtk.IconLookupFlags.FORCE_SIZE)

        return icon_info

    def make_device_icon(self, icon_info: Gtk.IconInfo, is_paired: bool = False, is_trusted: bool = False
                         ) -> cairo.Surface:
        window = self.get_window()
        scale = self.get_scale_factor()
        target = icon_info.load_surface(window)
        ctx = cairo.Context(target)

        if is_paired:
            _icon_info = self.get_icon_info("dialog-password", 16, False)
            assert _icon_info is not None
            paired_surface = _icon_info.load_surface(window)
            ctx.set_source_surface(paired_surface, 1 / scale, 1 / scale)
            ctx.paint_with_alpha(0.8)

        if is_trusted:
            _icon_info = self.get_icon_info("blueman-trust", 16, False)
            assert _icon_info is not None
            trusted_surface = _icon_info.load_surface(window)
            assert isinstance(target, cairo.ImageSurface)
            assert isinstance(trusted_surface, cairo.ImageSurface)
            height = target.get_height()
            mini_height = trusted_surface.get_height()
            y = height / scale - mini_height / scale - 1 / scale

            ctx.set_source_surface(trusted_surface, 1 / scale, y)
            ctx.paint_with_alpha(0.8)

        return target

    def device_remove_event(self, device: Device) -> None:
        tree_iter = self.find_device(device)
        assert tree_iter is not None

        row_fader = self.get(tree_iter, "row_fader")["row_fader"]
        super().device_remove_event(device)
        self.emit("device-selected", None, None)
        self._prepare_fader(row_fader).animate(start=row_fader.get_state(), end=0.0, duration=400)

    def device_add_event(self, device: Device) -> None:
        self.add_device(device)

    @staticmethod
    def make_caption(name: str, klass: str, address: str) -> str:
        return "<span size='x-large'>%(0)s</span>\n<span size='small'>%(1)s</span>\n<i>%(2)s</i>" \
               % {"0": html.escape(name), "1": klass, "2": address}

    @staticmethod
    def get_device_class(device: Device) -> str:
        klass = get_minor_class(device['Class'])
        if klass != _("Uncategorized"):
            return klass
        else:
            return get_major_class(device['Class'])

    def row_setup_event(self, tree_iter: Gtk.TreeIter, device: Device) -> None:
        if not self.get(tree_iter, "initial_anim")["initial_anim"]:
            model = self.props.model
            assert model is not None
            cell_fader = CellFade(self, model.get_path(tree_iter), [2, 3, 4, 5])
            row_fader = TreeRowFade(self, model.get_path(tree_iter))

            has_objpush = self._has_objpush(device)

            self.set(tree_iter, row_fader=row_fader, cell_fader=cell_fader, objpush=has_objpush)

            cell_fader.freeze()

            self._prepare_fader(row_fader).animate(start=0.0, end=1.0, duration=500)

            self.set(tree_iter, initial_anim=True)

        klass = get_minor_class(device['Class'])
        # Bluetooth >= 4 devices use Appearance property
        appearance = device["Appearance"]
        if klass != _("Uncategorized") and klass != _("Unknown"):
            description = klass
        elif klass == _("Unknown") and appearance:
            description = gatt_appearance_to_name(appearance)
        else:
            description = get_major_class(device['Class'])

        icon_info = self.get_icon_info(device["Icon"], 48, False)
        caption = self.make_caption(device['Alias'], description, device['Address'])

        self.set(tree_iter, caption=caption, icon_info=icon_info, alias=device['Alias'])

        try:
            self.row_update_event(tree_iter, "Trusted", device['Trusted'])
        except Exception as e:
            logging.exception(e)
        try:
            self.row_update_event(tree_iter, "Paired", device['Paired'])
        except Exception as e:
            logging.exception(e)
        try:
            self.row_update_event(tree_iter, "Connected", device["Connected"])
        except Exception as e:
            logging.exception(e)

        if device["Connected"]:
            self._monitor_power_levels(tree_iter, device)

    def _monitor_power_levels(self, tree_iter: Gtk.TreeIter, device: Device) -> None:
        if device["Address"] in self._monitored_devices:
            return

        assert self.Adapter is not None
        cinfo = conn_info(device["Address"], os.path.basename(self.Adapter.get_object_path()))
        try:
            cinfo.init()
        except ConnInfoReadError:
            logging.warning("Failed to get power levels, probably a LE device.")

        model = self.get_model()
        assert isinstance(model, Gtk.TreeModel)
        r = Gtk.TreeRowReference.new(model, model.get_path(tree_iter))
        self._update_power_levels(tree_iter, device, cinfo)
        GLib.timeout_add(1000, self._check_power_levels, r, cinfo, device["Address"])
        self._monitored_devices.add(device["Address"])

    def _check_power_levels(self, row_ref: Gtk.TreeRowReference, cinfo: conn_info, address: str) -> bool:
        if not row_ref.valid():
            logging.warning("stopping monitor (row does not exist)")
            cinfo.deinit()
            self._monitored_devices.remove(address)
            return False

        tree_iter = self.get_iter(row_ref.get_path())
        assert tree_iter is not None

        device = self.get(tree_iter, "device")["device"]

        if device["Connected"]:
            self._update_power_levels(tree_iter, device, cinfo)
            return True
        else:
            cinfo.deinit()
            self._disable_power_levels(tree_iter)
            self._monitored_devices.remove(address)
            return False

    def row_update_event(self, tree_iter: Gtk.TreeIter, key: str, value: Any) -> None:
        logging.info(f"{key} {value}")

        if key == "Trusted":
            if value:
                self.set(tree_iter, trusted=True)
            else:
                self.set(tree_iter, trusted=False)

        elif key == "Paired":
            if value:
                self.set(tree_iter, paired=True)
            else:
                self.set(tree_iter, paired=False)

        elif key == "Alias":
            device = self.get(tree_iter, "device")["device"]
            c = self.make_caption(value, self.get_device_class(device), device['Address'])
            self.set(tree_iter, caption=c, alias=value)

        elif key == "UUIDs":
            device = self.get(tree_iter, "device")["device"]
            has_objpush = self._has_objpush(device)
            self.set(tree_iter, objpush=has_objpush)

        elif key == "Connected":
            self.set(tree_iter, connected=value)

            if value:
                self._monitor_power_levels(tree_iter, self.get(tree_iter, "device")["device"])
            else:
                self._disable_power_levels(tree_iter)

    def _update_power_levels(self, tree_iter: Gtk.TreeIter, device: Device, cinfo: conn_info) -> None:
        row = self.get(tree_iter, "cell_fader", "battery", "rssi", "lq", "tpl")

        bars = {}

        if device["ServicesResolved"] and any(ServiceUUID(uuid).short_uuid == BATTERY_SERVICE_SVCLASS_ID
                                              for uuid in device["UUIDs"]):
            bars["battery"] = Battery(obj_path=device.get_object_path())["Percentage"]

        # cinfo init may fail for bluetooth devices version 4 and up
        # FIXME Workaround is horrible and we should show something better
        if cinfo.failed:
            if not bars:
                bars = {"rssi": 100.0, "tpl": 100.0, "lq": 100.0}
        else:
            try:
                bars["rssi"] = max(50 + float(cinfo.get_rssi()) / 127 * 50, 10)
            except ConnInfoReadError:
                bars["rssi"] = 50
            try:
                bars["lq"] = max(float(cinfo.get_lq()) / 255 * 100, 10)
            except ConnInfoReadError:
                bars["lq"] = 10
            try:
                bars["tpl"] = max(50 + float(cinfo.get_tpl()) / 127 * 50, 10)
            except ConnInfoReadError:
                bars["tpl"] = 50

        if row["battery"] == row["rssi"] == row["tpl"] == row["lq"] == 0:
            self._prepare_fader(row["cell_fader"]).animate(start=0.0, end=1.0, duration=400)

        w = 14 * self.get_scale_factor()
        h = 48 * self.get_scale_factor()

        for (name, perc) in bars.items():
            if round(row[name], -1) != round(perc, -1):
                icon_name = f"blueman-{name}-{int(round(perc, -1))}.png"
                icon = GdkPixbuf.Pixbuf.new_from_file_at_scale(os.path.join(PIXMAP_PATH, icon_name), w, h, True)
                self.set(tree_iter, **{name: perc, f"{name}_pb": icon})

    def _disable_power_levels(self, tree_iter: Gtk.TreeIter) -> None:
        row = self.get(tree_iter, "cell_fader", "battery", "rssi", "lq", "tpl")
        if row["battery"] == row["rssi"] == row["tpl"] == row["lq"] == 0:
            return

        self.set(tree_iter, rssi=0, lq=0, tpl=0)
        self._prepare_fader(row["cell_fader"], lambda: self.set(tree_iter, rssi_pb=None, lq_pb=None, tpl_pb=None))\
            .animate(start=1.0, end=0.0, duration=400)

    def _prepare_fader(self, fader: AnimBase, callback: Optional[Callable[[], None]] = None) -> AnimBase:
        def on_finished(finished_fader: AnimBase) -> None:
            finished_fader.disconnect(handler)
            finished_fader.freeze()
            if callback:
                callback()

        fader.thaw()
        handler = fader.connect("animation-finished", on_finished)
        return fader

    def tooltip_query(self, _tw: Gtk.Widget, x: int, y: int, _kb: bool, tooltip: Gtk.Tooltip) -> bool:
        path = self.get_path_at_pos(x, y)
        if path is None:
            return False

        if path[0] != self.tooltip_row or path[1] != self.tooltip_col:
            self.tooltip_row = path[0]
            self.tooltip_col = path[1]
            return False

        if path[1] == self.columns["device_surface"]:
            tree_iter = self.get_iter(path[0])
            assert tree_iter is not None

            row = self.get(tree_iter, "trusted", "paired")
            trusted = row["trusted"]
            paired = row["paired"]
            if trusted and paired:
                tooltip.set_markup(_("<b>Trusted and Paired</b>"))
            elif paired:
                tooltip.set_markup(_("<b>Paired</b>"))
            elif trusted:
                tooltip.set_markup(_("<b>Trusted</b>"))
            else:
                return False

            self.tooltip_row = path[0]
            self.tooltip_col = path[1]
            return True

        elif path[1] == self.columns["battery_pb"] \
                or path[1] == self.columns["tpl_pb"] \
                or path[1] == self.columns["lq_pb"] \
                or path[1] == self.columns["rssi_pb"]:
            tree_iter = self.get_iter(path[0])
            assert tree_iter is not None

            dt = self.get(tree_iter, "connected")["connected"]
            if not dt:
                return False

            lines = [_("<b>Connected</b>")]

            battery = self.get(tree_iter, "battery")["battery"]
            rssi = self.get(tree_iter, "rssi")["rssi"]
            lq = self.get(tree_iter, "lq")["lq"]
            tpl = self.get(tree_iter, "tpl")["tpl"]

            if battery != 0:
                if path[1] == self.columns["battery_pb"]:
                    lines.append(f"<b>Battery: {int(battery)}%</b>")
                else:
                    lines.append(f"Battery: {int(battery)}%")

            if rssi != 0:
                if rssi < 30:
                    rssi_state = _("Poor")
                elif rssi < 40:
                    rssi_state = _("Sub-optimal")
                elif rssi < 60:
                    rssi_state = _("Optimal")
                elif rssi < 70:
                    rssi_state = _("Much")
                else:
                    rssi_state = _("Too much")

                if path[1] == self.columns["rssi_pb"]:
                    lines.append(_("<b>Received Signal Strength: %(rssi)u%%</b> <i>(%(rssi_state)s)</i>") %
                                 {"rssi": rssi, "rssi_state": rssi_state})
                else:
                    lines.append(_("Received Signal Strength: %(rssi)u%% <i>(%(rssi_state)s)</i>") %
                                 {"rssi": rssi, "rssi_state": rssi_state})

            if lq != 0:
                if path[1] == self.columns["lq_pb"]:
                    lines.append(_("<b>Link Quality: %(lq)u%%</b>") % {"lq": lq})
                else:
                    lines.append(_("Link Quality: %(lq)u%%") % {"lq": lq})

            if tpl != 0:
                if tpl < 30:
                    tpl_state = _("Low")
                elif tpl < 40:
                    tpl_state = _("Sub-optimal")
                elif tpl < 60:
                    tpl_state = _("Optimal")
                elif tpl < 70:
                    tpl_state = _("High")
                else:
                    tpl_state = _("Very High")

                if path[1] == self.columns["tpl_pb"]:
                    lines.append(_("<b>Transmit Power Level: %(tpl)u%%</b> <i>(%(tpl_state)s)</i>") %
                                 {"tpl": tpl, "tpl_state": tpl_state})
                else:
                    lines.append(_("Transmit Power Level: %(tpl)u%% <i>(%(tpl_state)s)</i>") %
                                 {"tpl": tpl, "tpl_state": tpl_state})

            tooltip.set_markup("\n".join(lines))
            self.tooltip_row = path[0]
            self.tooltip_col = path[1]
            return True
        return False

    def _has_objpush(self, device: Device) -> bool:
        if device is None:
            return False

        for uuid in device["UUIDs"]:
            if ServiceUUID(uuid).short_uuid == OBEX_OBJPUSH_SVCLASS_ID:
                return True
        return False

    def _set_cell_data(self, _col: Gtk.TreeViewColumn, cell: Gtk.CellRenderer, _model: Gtk.TreeModel,
                       tree_iter: Gtk.TreeIter, data: Optional[str]) -> None:
        if data is None:
            row = self.get(tree_iter, "icon_info", "trusted", "paired")
            surface = self.make_device_icon(row["icon_info"], row["paired"], row["trusted"])
            cell.set_property("surface", surface)
        else:
            window = self.get_window()
            scale = self.get_scale_factor()
            pb = self.get(tree_iter, data + "_pb")[data + "_pb"]
            if pb:
                surface = Gdk.cairo_surface_create_from_pixbuf(pb, scale, window)
                cell.set_property("surface", surface)
            else:
                cell.set_property("surface", None)

    def add_device(self, device: Device) -> None:
        if "Name" not in device and self.Config["hide-unnamed"]:
            logging.info(f"Hiding unnamed device: {device['Address']}")
            return

        super().add_device(device)



from gettext import gettext as _
from operator import itemgetter
import time
import logging
from typing import Dict, List, TYPE_CHECKING, Optional, Callable, cast, Union

from blueman.bluez.Device import Device
from blueman.bluez.errors import DBusNoSuchAdapterError
from blueman.gui.Notification import Notification
from blueman.Sdp import ServiceUUID
from blueman.plugins.AppletPlugin import AppletPlugin
from blueman.plugins.applet.PowerManager import PowerManager, PowerStateListener

if TYPE_CHECKING:
    from blueman.plugins.applet.Menu import MenuItemDict

    from typing_extensions import TypedDict

    class _ItemBase(TypedDict):
        adapter: str
        address: str
        alias: str
        icon: str
        name: str
        uuid: str

    class Item(_ItemBase):
        time: float
        device: Optional[str]
        mitem: Optional[MenuItemDict]

    class StoredIcon(_ItemBase):
        time: str

REGISTRY_VERSION = 0


class RecentConns(AppletPlugin, PowerStateListener):
    __depends__ = ["DBusService", "Menu"]
    __icon__ = "document-open-recent"
    __description__ = _("Provides a menu item that contains last used connections for quick access")
    __author__ = "Walmis"

    __gsettings__ = {
        "schema": "org.blueman.plugins.recentconns",
        "path": None
    }
    __options__ = {
        "max-items": {"type": int,
                      "default": 6,
                      # the maximum number of items RecentConns menu will display
                      "name": _("Maximum items"),
                      "desc": _("The maximum number of items recent connections menu will display."),
                      "range": (6, 20)},
        "recent-connections": {"type": list, "default": "[]"}
    }

    _items = None

    def on_load(self) -> None:
        self._adapters: Dict[str, str] = {}
        self.__menuitems: List["MenuItemDict"] = []

        self.item = self.parent.Plugins.Menu.add(self, 52, text=_("Recent _Connections") + "…",
                                                 icon_name="document-open-recent", submenu_function=self.get_menu)
        self.parent.Plugins.Menu.add(self, 53)

        self.deferred = False

    def _store_state(self) -> None:
        self.set_option("recent-connections", [
            {
                "adapter": item["adapter"],
                "address": item["address"],
                "alias": item["alias"],
                "icon": item["icon"],
                "name": item["name"],
                "uuid": item["uuid"],
                "time": str(item["time"]),
            } for item in self.items
        ])

    def change_sensitivity(self, sensitive: bool) -> None:
        if 'PowerManager' in self.parent.Plugins.get_loaded():
            power = self.parent.Plugins.PowerManager.get_bluetooth_status()
        else:
            power = True

        sensitive = sensitive and \
            self.parent.Manager is not None and \
            power and \
            self.items is not None and \
            (len(self.items) > 0)

        self.item.set_sensitive(sensitive)

    def on_power_state_changed(self, manager: PowerManager, state: bool) -> None:
        self.change_sensitivity(state)
        if state and self.deferred:
            self.deferred = False
            self.on_manager_state_changed(state)

    def on_unload(self) -> None:
        self.parent.Plugins.Menu.unregister(self)

    def initialize(self) -> None:
        logging.info("rebuilding menu")

        self.__menuitems = []
        self.parent.Plugins.Menu.on_menu_changed()

        self.items.sort(key=itemgetter("time"), reverse=True)

        self._items = self.items[0:self.get_option("max-items")]
        self.items.reverse()

        if len(self.items) == 0:
            self.change_sensitivity(False)
        else:
            self.change_sensitivity(True)

        count = 0
        for item in self.items:
            if count < self.get_option("max-items"):
                self.add_item(item)
                count += 1

    def on_manager_state_changed(self, state: bool) -> None:
        if state:
            if 'PowerManager' in self.parent.Plugins.get_loaded():
                if not self.parent.Plugins.PowerManager.get_bluetooth_status():
                    self.deferred = True
                    self.item.set_sensitive(False)
                    return

            self.item.set_sensitive(True)
            adapters = self.parent.Manager.get_adapters()

            self._adapters = {adapter.get_object_path(): adapter["Address"]
                              for adapter in adapters}

            for i in self.items:
                device_path = self._get_device_path(i["adapter"], i["address"])
                if device_path:
                    i["device"] = device_path

            self.initialize()
        else:
            self.item.set_sensitive(False)
            return

        self.change_sensitivity(state)

    def on_device_created(self, path: str) -> None:
        self._items = None
        self.initialize()

    def on_device_removed(self, path: str) -> None:
        for item in reversed(self.items):
            if item['device'] == path:
                self.items.remove(item)
                self.initialize()

    def on_adapter_added(self, path: str) -> None:
        a = self.parent.Manager.get_adapter(path)
        self._adapters[path] = a["Address"]
        self.initialize()

    def on_adapter_removed(self, path: str) -> None:
        if path in self._adapters:
            del self._adapters[path]
        else:
            logging.warning("Adapter not found in list")

        self.initialize()

    def notify(self, object_path: str, uuid: str) -> None:
        device = Device(obj_path=object_path)
        logging.info(f"{device} {uuid}")
        try:
            adapter = self.parent.Manager.get_adapter(device['Adapter'])
        except DBusNoSuchAdapterError:
            logging.warning("adapter not found")
            return

        item: "Item" = {
            "adapter": adapter["Address"],
            "address": device['Address'],
            "alias": device['Alias'],
            "icon": device['Icon'],
            "name": ServiceUUID(uuid).name,
            "uuid": uuid,
            "time": time.time(),
            "device": object_path,
            "mitem": None
        }

        for i in self.items:
            if i["adapter"] == item["adapter"] and \
                    i["address"] == item["address"] and \
                    i["uuid"] == item["uuid"]:
                i["time"] = item["time"]

                i["device"] = item["device"]
                self.initialize()
                return

        self.items.append(item)
        self.initialize()

        self._store_state()

    def on_item_activated(self, item: "Item") -> None:
        logging.info(f"Connect {item['address']} {item['uuid']}")

        assert item["mitem"] is not None

        item["mitem"]["sensitive"] = False
        self.parent.Plugins.Menu.on_menu_changed()

        def reply() -> None:
            assert item["mitem"] is not None  # https://github.com/python/mypy/issues/2608
            Notification(_("Connected"), _("Connected to %s") % item["mitem"]["text"],
                         icon_name=item["icon"]).show()
            item["mitem"]["sensitive"] = True
            self.parent.Plugins.Menu.on_menu_changed()

        def err(reason: Union[Exception, str]) -> None:
            Notification(_("Failed to connect"), str(reason).split(": ")[-1],
                         icon_name="dialog-error").show()
            assert item["mitem"] is not None  # https://github.com/python/mypy/issues/2608
            item["mitem"]["sensitive"] = True
            self.parent.Plugins.Menu.on_menu_changed()

        self.parent.Plugins.DBusService.connect_service(item["device"], item["uuid"], reply, err)

    def add_item(self, item: "Item") -> None:
        if item["adapter"] not in self._adapters.values():
            item["device"] = None
        elif not item["device"] and item["adapter"] in self._adapters.values():
            device = self._get_device_path(item["adapter"], (item["address"]))
            if device is not None:
                item["device"] = device
            else:
                self.items.remove(item)
                self.initialize()

        mitem: "MenuItemDict" = {
            "text": _("%(service)s on %(device)s") % {"service": item["name"], "device": item["alias"]},
            "markup": True,
            "icon_name": item["mitem"]["icon_name"] if item["mitem"] is not None else item["icon"],
            "sensitive": item["device"] is not None,
            "tooltip": None if item["device"] is None else _("Adapter for this connection is not available"),
            "callback": (item["mitem"]["callback"] if item["mitem"] is not None
                         else cast(Callable[[], None], lambda itm=item: self.on_item_activated(itm)))
        }

        item["mitem"] = mitem

        self.__menuitems.append(mitem)
        self.parent.Plugins.Menu.on_menu_changed()

    def get_menu(self) -> List["MenuItemDict"]:
        return self.__menuitems

    def _get_device_path(self, adapter_path: str, address: str) -> Optional[str]:
        try:
            adapter = self.parent.Manager.get_adapter(adapter_path)
        except DBusNoSuchAdapterError:
            return None

        device = self.parent.Manager.find_device(address, adapter.get_object_path())
        return device.get_object_path() if device is not None else None

    @property
    def items(self) -> List["Item"]:
        if self._items is not None:
            return self._items

        items = self.get_option("recent-connections")

        if not items:
            self._items = []
            return self._items

        self._items = [
            {
                "adapter": i["adapter"],
                "address": i["address"],
                "alias": i["alias"],
                "icon": i["icon"],
                "name": i["name"],
                "uuid": i["uuid"],
                "time": float(i["time"]),
                "device": (self._get_device_path(i["adapter"], i["address"])
                           if i["adapter"] in self._adapters.values() else None),
                "mitem": None
            }
            for i in items
            if i["adapter"] not in self._adapters.values() or self._get_device_path(i["adapter"], i["address"])
        ]

        return self._items



import logging
from typing import List, Optional

from blueman.bluez.obex.Base import Base
from gi.repository import GObject, GLib

from blueman.bluemantyping import GSignals


class Transfer(Base):
    __gsignals__: GSignals = {
        'progress': (GObject.SignalFlags.NO_HOOKS, None, (int,)),
        'completed': (GObject.SignalFlags.NO_HOOKS, None, ()),
        'error': (GObject.SignalFlags.NO_HOOKS, None, ())
    }

    _interface_name = 'org.bluez.obex.Transfer1'

    def __init__(self, obj_path: str):
        super().__init__(obj_path=obj_path)

    @property
    def filename(self) -> Optional[str]:
        name: Optional[str] = self.get("Filename")
        return name

    @property
    def name(self) -> str:
        name: str = self.get("Name")
        return name

    @property
    def session(self) -> str:
        session: str = self.get("Session")
        return session

    @property
    def size(self) -> Optional[int]:
        size: Optional[int] = self.get("Size")
        return size

    def do_g_properties_changed(self, changed_properties: GLib.Variant, _invalidated_properties: List[str]) -> None:
        for name, value in changed_properties.unpack().items():
            logging.debug(f"{self.get_object_path()} {name} {value}")
            if name == 'Transferred':
                self.emit('progress', value)
            elif name == 'Status':
                if value == 'complete':
                    self.emit('completed')
                elif value == 'error':
                    self.emit('error')



from gettext import gettext as _
from typing import List, Tuple, Iterable, Callable, Any, Optional

from gi.repository import Gtk, Gdk

import logging

from gi.repository.GObject import GObject

from blueman.Functions import create_menuitem
from blueman.Sdp import ServiceUUID
from blueman.bluez.Device import Device
from blueman.bluez.errors import BluezDBusException
from blueman.gui.manager.ManagerDeviceMenu import MenuItemsProvider, ManagerDeviceMenu, DeviceMenuItem

from blueman.plugins.ManagerPlugin import ManagerPlugin


def show_info(device: Device, parent: Gtk.Window) -> None:
    def format_boolean(x: bool) -> str:
        return _('yes') if x else _('no')

    def format_rssi(rssi: int) -> str:
        if rssi in [0x99, 0x7f]:
            return f'invalid (0x{rssi:02x})'
        else:
            return f'{rssi} dBm (0x{rssi:02x})'

    def format_uuids(uuids: Iterable[str]) -> str:
        return "\n".join([uuid + ' ' + ServiceUUID(uuid).name for uuid in uuids])

    store = Gtk.ListStore(str, str)
    view = Gtk.TreeView(model=store, headers_visible=False)
    view_selection = view.get_selection()
    view_selection.set_mode(Gtk.SelectionMode.MULTIPLE)

    def on_accel_activated(_group: Gtk.AccelGroup, _dialog: GObject, key: int, _modifier: Gdk.ModifierType) -> bool:
        if key != 99:
            logging.warning(f"Ignoring key {key}")
            return False

        store, paths = view_selection.get_selected_rows()

        text = []
        for path in paths:
            row = store[path]
            text.append(row[-1])

        logging.info("\n".join(text))
        clipboard.set_text("\n".join(text), -1)

        return False

    clipboard = Gtk.Clipboard.get(Gdk.SELECTION_CLIPBOARD)
    dialog = Gtk.Dialog(icon_name="blueman", title="blueman")
    dialog.set_transient_for(parent)
    dialog_content_area = dialog.get_content_area()

    label = Gtk.Label()
    label.set_markup(_("<big>Select row(s) and use <i>Control + C</i> to copy</big>"))
    label.show()
    dialog_content_area.pack_start(label, True, False, 0)

    accelgroup = Gtk.AccelGroup()
    dialog.add_accel_group(accelgroup)

    key, mod = Gtk.accelerator_parse("<Control>C")
    accelgroup.connect(key, mod, Gtk.AccelFlags.MASK, on_accel_activated)

    for i in range(2):
        column = Gtk.TreeViewColumn()
        cell = Gtk.CellRendererText()
        column.pack_start(cell, True)
        column.add_attribute(cell, 'text', i)
        view.append_column(column)
    dialog_content_area.pack_start(view, True, False, 0)
    view.show_all()

    properties: Iterable[Tuple[str, Optional[Callable[[Any], str]]]] = (
        ('Address', None),
        ('AddressType', None),
        ('Name', None),
        ('Alias', None),
        ('Class', lambda x: f"0x{x:06x}"),
        ('Appearance', lambda x: f"0x{x:04x}"),
        ('Icon', None),
        ('Paired', format_boolean),
        ('Trusted', format_boolean),
        ('Blocked', format_boolean),
        ('LegacyPairing', format_boolean),
        ('RSSI', format_rssi),
        ('Connected', format_boolean),
        ('UUIDs', format_uuids),
        ('Modalias', None),
        ('Adapter', None),
        # FIXME below 3 we need some sample data to decode and display properly
        ('ManufacturerData', str),
        ('ServiceData', str),
        ('AdvertisingData', str)
    )
    for name, func in properties:
        try:
            if func is None:
                store.append((name, device.get(name)))
            else:
                store.append((name, func(device.get(name))))
        except BluezDBusException:
            logging.info(f"Could not get property {name}")
            pass
        except ValueError:
            logging.info(f"Could not add property {name}")
            pass

    dialog.run()
    dialog.destroy()


class Info(ManagerPlugin, MenuItemsProvider):
    def on_request_menu_items(self, manager_menu: ManagerDeviceMenu, device: Device) -> List[DeviceMenuItem]:
        item = create_menuitem(_("_Info"), "dialog-information")
        item.props.tooltip_text = _("Show device information")
        _window = manager_menu.get_toplevel()
        assert isinstance(_window, Gtk.Window)
        window = _window  # https://github.com/python/mypy/issues/2608
        item.connect('activate', lambda x: show_info(device, window))
        return [DeviceMenuItem(item, DeviceMenuItem.Group.ACTIONS, 400)]



from gettext import gettext as _
from typing import Any

from blueman.Sdp import ServiceUUID, BATTERY_SERVICE_SVCLASS_ID
from blueman.bluez.Battery import Battery as BluezBattery
from blueman.bluez.Device import Device
from blueman.gui.Notification import Notification
from blueman.plugins.AppletPlugin import AppletPlugin


class Battery(AppletPlugin):
    __author__ = "cschramm"
    __icon__ = "battery"
    __description__ = _("Shows desktop notifications with battery percentage when devices get connected.")

    def on_device_property_changed(self, path: str, key: str, value: Any) -> None:
        if key == "ServicesResolved" and value:
            device = Device(obj_path=path)
            if self.applicable(device):
                text = "%d%%" % BluezBattery(obj_path=path)["Percentage"]
                Notification(device["Alias"], text, icon_name="battery").show()

    @staticmethod
    def applicable(device: Device) -> bool:
        return any(ServiceUUID(uuid).short_uuid == BATTERY_SERVICE_SVCLASS_ID for uuid in device["UUIDs"])



from importlib import import_module
import logging
import os
import sys
from blueman.main.DBusProxies import AppletService
from gi.repository import Gio, GLib


class BluemanTray(Gio.Application):
    def __init__(self) -> None:
        super().__init__(application_id="org.blueman.Tray", flags=Gio.ApplicationFlags.FLAGS_NONE)
        self._active = False

    def do_activate(self) -> None:
        if self._active:
            logging.info("Already running, restarting instance")
            os.execv(sys.argv[0], sys.argv)

        Gio.bus_watch_name(Gio.BusType.SESSION, 'org.blueman.Applet', Gio.BusNameWatcherFlags.NONE,
                           self._on_name_appeared, self._on_name_vanished)
        self.hold()

    def _on_name_appeared(self, _connection: Gio.DBusConnection, name: str, _owner: str) -> None:
        logging.debug("Applet started on name %s, showing indicator" % name)

        applet = AppletService()
        indicator_name = applet.GetStatusIconImplementation()
        logging.info(f'Using indicator "{indicator_name}"')
        indicator_class = getattr(import_module('blueman.main.indicators.' + indicator_name), indicator_name)
        self.indicator = indicator_class(applet.GetIconName(), self._activate_menu_item, self._activate_status_icon)

        applet.connect('g-signal', self.on_signal)

        self.indicator.set_text(applet.GetText())
        self.indicator.set_visibility(applet.GetVisibility())
        self.indicator.set_menu(applet.GetMenu())

        self._active = True

    def _on_name_vanished(self, _connection: Gio.DBusConnection, _name: str) -> None:
        logging.debug("Applet shutdown or not available at startup")
        self.quit()

    def _activate_menu_item(self, *indexes: int) -> None:
        AppletService().ActivateMenuItem('(ai)', indexes)

    def _activate_status_icon(self) -> None:
        AppletService().Activate()

    def on_signal(self, _applet: AppletService, _sender_name: str, signal_name: str, args: GLib.Variant) -> None:
        if signal_name == 'IconNameChanged':
            self.indicator.set_icon(*args)
        elif signal_name == 'TextChanged':
            self.indicator.set_text(*args)
        elif signal_name == 'VisibilityChanged':
            self.indicator.set_visibility(*args)
        elif signal_name == 'MenuChanged':
            self.indicator.set_menu(*args)



from blueman.bluez.Base import Base
from gi.repository import GLib


class AgentManager(Base):
    _interface_name = 'org.bluez.AgentManager1'
    _obj_path = '/org/bluez'

    def __init__(self) -> None:
        super().__init__(obj_path=self._obj_path)

    def register_agent(self, agent_path: str, capability: str = "", default: bool = False) -> None:
        param = GLib.Variant('(os)', (agent_path, capability))
        self._call('RegisterAgent', param)
        if default:
            default_param = GLib.Variant('(o)', (agent_path,))
            self._call('RequestDefaultAgent', default_param)

    def unregister_agent(self, agent_path: str) -> None:
        param = GLib.Variant('(o)', (agent_path,))
        self._call('UnregisterAgent', param)



from gettext import gettext as _
import logging
from typing import TYPE_CHECKING, List, Type, Dict

from blueman.gui.GenericList import GenericList, ListDataDict
from blueman.main.Builder import Builder
from blueman.main.PluginManager import PluginManager
from blueman.plugins.AppletPlugin import AppletPlugin
from blueman.plugins.BasePlugin import Option

import gi
gi.require_version("Gtk", "3.0")
from gi.repository import Gtk
gi.require_version("Gdk", "3.0")
from gi.repository import Gdk

if TYPE_CHECKING:
    from blueman.main.Applet import BluemanApplet


class SettingsWidget(Gtk.Box):
    def __init__(self, inst: AppletPlugin, orientation: Gtk.Orientation = Gtk.Orientation.VERTICAL) -> None:
        super().__init__(
            name="SettingsWidget",
            orientation=orientation,
            spacing=2
        )
        self.inst = inst

        self.construct_settings()
        self.show_all()

    def construct_settings(self) -> None:
        for k, v in self.inst.__class__.__options__.items():
            if len(v) > 2:

                label = Gtk.Label(label=v["name"])
                label.props.xalign = 0.0

                w = self.get_control_widget(k, v)

                self.pack_start(w, False, False, 0)

                label = Gtk.Label(label="<i >" + v["desc"] + "</i>", wrap=True, use_markup=True, xalign=0.0)
                self.pack_start(label, False, False, 0)

    def handle_change(self, widget: Gtk.Widget, opt: str, params: "Option", prop: str) -> None:
        val = params["type"](getattr(widget.props, prop))
        logging.debug(f"changed {opt} {val}")

        self.inst.set_option(opt, val)

    def get_control_widget(self, opt: str, params: "Option") -> Gtk.Widget:
        if params["type"] == bool:
            c = Gtk.CheckButton(label=params["name"])

            c.props.active = self.inst.get_option(opt)
            c.connect("toggled", self.handle_change, opt, params, "active")
            return c

        elif params["type"] == int:
            b = Gtk.Box(spacing=6)
            label = Gtk.Label(label=params["name"])
            b.pack_start(label, False, False, 0)

            r = Gtk.SpinButton(numeric=True)
            b.pack_start(r, False, False, 6)

            r.set_increments(1, 3)
            r.set_range(params["range"][0], params["range"][1])

            r.props.value = self.inst.get_option(opt)
            r.connect("value-changed", self.handle_change, opt, params, "value")

            return b

        elif params["type"] == str:
            b = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL, spacing=6)
            label = Gtk.Label(label=params["name"])
            b.pack_start(label, False, False, 0)

            e = Gtk.Entry()
            b.pack_start(e, False, False, 6)

            e.props.text = self.inst.get_option(opt)
            e.connect("changed", self.handle_change, opt, params, "text")

            return b
        raise ValueError()


class PluginDialog(Gtk.Window):
    def __init__(self, applet: "BluemanApplet") -> None:
        super().__init__(
            title=_("Plugins"),
            icon_name="blueman",
            name="PluginDialog",
            border_width=6,
            default_width=490,
            default_height=380,
            resizable=False,
            visible=False
        )

        self.applet = applet

        builder = Builder("applet-plugins-widget.ui")

        self.description = builder.get_widget("description", Gtk.Label)

        self.icon = builder.get_widget("icon", Gtk.Image)
        self.author_txt = builder.get_widget("author_txt", Gtk.Label)
        self.depends_hdr = builder.get_widget("depends_hdr", Gtk.Widget)
        self.depends_txt = builder.get_widget("depends_txt", Gtk.Label)
        self.conflicts_hdr = builder.get_widget("conflicts_hdr", Gtk.Widget)
        self.conflicts_txt = builder.get_widget("conflicts_txt", Gtk.Label)
        self.plugin_name = builder.get_widget("name", Gtk.Label)

        self.main_container = builder.get_widget("main_container", Gtk.Bin)
        self.content_grid = builder.get_widget("content", Gtk.Widget)

        self.b_prefs = builder.get_widget("b_prefs", Gtk.ToggleButton)
        self.b_prefs.connect("toggled", self.on_prefs_toggled)

        self.add(builder.get_widget("all", Gtk.Container))

        cr = Gtk.CellRendererToggle()
        cr.connect("toggled", self.on_toggled)

        data: List[ListDataDict] = [
            {"id": "active", "type": bool, "renderer": cr, "render_attrs": {"active": 0, "activatable": 1,
                                                                            "visible": 1}},
            {"id": "activatable", "type": bool},
            {"id": "icon", "type": str, "renderer": Gtk.CellRendererPixbuf(), "render_attrs": {"icon-name": 2}},
            # device caption
            {"id": "desc", "type": str, "renderer": Gtk.CellRendererText(), "render_attrs": {"markup": 3},
             "view_props": {"expand": True}},
            {"id": "name", "type": str},
        ]

        self.list = GenericList(data, headers_visible=False, visible=True)
        self.list.liststore.set_sort_column_id(3, Gtk.SortType.ASCENDING)
        self.list.liststore.set_sort_func(3, self.list_compare_func)

        self.list.selection.connect("changed", self.on_selection_changed)

        plugin_list = builder.get_widget("plugin_list", Gtk.ScrolledWindow)
        plugin_info = builder.get_widget("main_scrolled_window", Gtk.ScrolledWindow)
        plugin_list.add(self.list)

        # Disable overlay scrolling
        if Gtk.get_minor_version() >= 16:
            plugin_list.props.overlay_scrolling = False
            plugin_info.props.overlay_scrolling = False

        self.populate()

        self.sig_a: int = self.applet.Plugins.connect("plugin-loaded", self.plugin_state_changed, True)
        self.sig_b: int = self.applet.Plugins.connect("plugin-unloaded", self.plugin_state_changed, False)
        self.connect("delete-event", self._on_close)

        self.list.set_cursor(0)

    def list_compare_func(self, _treemodel: Gtk.TreeModel, iter1: Gtk.TreeIter, iter2: Gtk.TreeIter, _user_data: object
                          ) -> int:
        a = self.list.get(iter1, "activatable", "name")
        b = self.list.get(iter2, "activatable", "name")

        if (a["activatable"] and b["activatable"]) or (not a["activatable"] and not b["activatable"]):
            if a["name"] == b["name"]:
                return 0
            if a["name"] < b["name"]:
                return -1
            else:
                return 1

        else:
            if a["activatable"] and not b["activatable"]:
                return -1
            elif not a["activatable"] and b["activatable"]:
                return 1
            else:
                return 0

    def _on_close(self, _widget: Gtk.Widget, _event: Gdk.Event) -> bool:
        self.applet.Plugins.disconnect(self.sig_a)
        self.applet.Plugins.disconnect(self.sig_b)
        return False

    def on_selection_changed(self, selection: Gtk.TreeSelection) -> None:
        model, tree_iter = selection.get_selected()
        assert tree_iter is not None

        name = self.list.get(tree_iter, "name")["name"]
        cls: Type[AppletPlugin] = self.applet.Plugins.get_classes()[name]
        self.plugin_name.props.label = "<b>" + name + "</b>"
        self.icon.props.icon_name = cls.__icon__
        self.author_txt.props.label = cls.__author__
        self.description.props.label = cls.__description__

        if cls.__depends__:
            self.depends_hdr.props.visible = True
            self.depends_txt.props.visible = True
            self.depends_txt.props.label = ", ".join(cls.__depends__)
        else:
            self.depends_hdr.props.visible = False
            self.depends_txt.props.visible = False

        if cls.__conflicts__:
            self.conflicts_hdr.props.visible = True
            self.conflicts_txt.props.visible = True
            self.conflicts_txt.props.label = ", ".join(cls.__conflicts__)
        else:
            self.conflicts_hdr.props.visible = False
            self.conflicts_txt.props.visible = False

        if cls.is_configurable() and name in self.applet.Plugins.get_loaded():
            self.b_prefs.props.sensitive = True
        else:
            self.b_prefs.props.sensitive = False

        self.update_config_widget(cls)

    def on_prefs_toggled(self, _button: Gtk.ToggleButton) -> None:
        model, tree_iter = self.list.selection.get_selected()
        assert tree_iter is not None
        name = self.list.get(tree_iter, "name")["name"]
        cls: Type[AppletPlugin] = self.applet.Plugins.get_classes()[name]

        self.update_config_widget(cls)

    def update_config_widget(self, cls: Type[AppletPlugin]) -> None:
        if self.b_prefs.props.active:
            if not cls.is_configurable():
                self.b_prefs.props.active = False
                return

            inst = cls.get_instance()
            if not inst:
                self.b_prefs.props.active = False
            else:
                c = self.main_container.get_child()
                assert c is not None
                self.main_container.remove(c)
                if isinstance(c, SettingsWidget):
                    c.destroy()

                self.main_container.add(SettingsWidget(inst))

        else:
            c = self.main_container.get_child()
            assert c is not None
            self.main_container.remove(c)
            if isinstance(c, SettingsWidget):
                c.destroy()
            self.main_container.add(self.content_grid)

    def populate(self) -> None:
        classes: Dict[str, Type[AppletPlugin]] = self.applet.Plugins.get_classes()
        loaded = self.applet.Plugins.get_loaded()
        for name, cls in classes.items():
            if cls.is_configurable():
                desc = f"<span weight=\"bold\">{name}</span>"
            else:
                desc = name
            self.list.append(active=(name in loaded), icon=cls.__icon__, activatable=cls.__unloadable__, name=name,
                             desc=desc)

    def plugin_state_changed(self, _plugins: PluginManager, name: str, loaded: bool) -> None:
        row = self.list.get_conditional(name=name)
        self.list.set(row[0], active=loaded)

        cls: Type[AppletPlugin] = self.applet.Plugins.get_classes()[name]
        if not loaded:
            self.update_config_widget(cls)
            self.b_prefs.props.sensitive = False
        elif cls.is_configurable():
            self.b_prefs.props.sensitive = True

    def on_toggled(self, _toggle: Gtk.CellRendererToggle, path: str) -> None:
        name = self.list.get(path, "name")["name"]

        deps = self.applet.Plugins.get_dependencies()[name]
        loaded = self.applet.Plugins.get_loaded()
        to_unload = []
        for dep in deps:
            if dep in loaded:
                to_unload.append(dep)

        if to_unload:
            dialog = Gtk.MessageDialog(parent=self, type=Gtk.MessageType.QUESTION, buttons=Gtk.ButtonsType.YES_NO)
            dialog.props.secondary_use_markup = True
            dialog.props.icon_name = "blueman"
            dialog.props.text = _("Dependency issue")
            dialog.props.secondary_text = \
                _("Plugin <b>\"%(0)s\"</b> depends on <b>%(1)s</b>. Unloading <b>%(1)s</b> will also unload <b>"
                  "\"%(0)s\"</b>.\nProceed?") % {"0": ", ".join(to_unload), "1": name}

            resp = dialog.run()
            if resp != Gtk.ResponseType.YES:
                dialog.destroy()
                return

            dialog.destroy()

        conflicts = self.applet.Plugins.get_conflicts()[name]
        to_unload = []
        for conf in conflicts:
            if conf in loaded:
                to_unload.append(conf)

        if to_unload:
            dialog = Gtk.MessageDialog(parent=self, type=Gtk.MessageType.QUESTION, buttons=Gtk.ButtonsType.YES_NO)
            dialog.props.secondary_use_markup = True
            dialog.props.icon_name = "blueman"
            dialog.props.text = _("Dependency issue")
            dialog.props.secondary_text = \
                _("Plugin <b>%(0)s</b> conflicts with <b>%(1)s</b>. Loading <b>%(1)s</b> will unload <b>%(0)s</b>."
                  "\nProceed?") % {"0": ", ".join(to_unload), "1": name}

            resp = dialog.run()
            if resp != Gtk.ResponseType.YES:
                dialog.destroy()
                return

            dialog.destroy()

            for p in to_unload:
                self.applet.Plugins.set_config(p, False)

        self.applet.Plugins.set_config(name, name not in self.applet.Plugins.get_loaded())
